@@ -12,29 +12,6 @@
 #ifndef _confmagic_h_
 #define _confmagic_h_
 
-#define inline __inline
-
-/*
- * (which isn't exportable from the U.S.), then don't encrypt
- */
-#ifndef HAS_CRYPT
-#define crypt(s,t) (s)
-#endif
-
-/* You better get with the 90's if this isn't true! */
-#define HAS_IEEE_MATH
-
-#ifndef HAVE_SIGCHLD
-#define SIGCHLD	SIGCLD
-#elif !defined(HAVE_SIGCLD)
-#define SIGCLD	SIGCHLD
-#endif
-
-#ifndef HAVE_STRCOLL
-#undef strcoll
-#define strcoll strcmp
-#endif
-
 #ifndef HAVE_SNPRINTF
 #ifdef HAVE__VSNPRINTF_S
 #define snprintf sane_snprintf_s
@@ -56,10 +33,4 @@ int sane_snprintf_s(char *, size_t, const char *, ...);
 #endif
 #endif
 
-#if defined(HAVE_POLLTS) && !defined(HAVE_PPOLL)
-/* Linux's ppoll() is identical to NetBSD's pollts() in all but name. */
-#define ppoll pollts
-#define HAVE_PPOLL
-#endif
-
 #endif
