@@ -7,6 +7,12 @@
 #ifndef INTMAP_H
 #define INTMAP_H
 
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif /* HAVE_STDINT_H */
+
+#include "mushtype.h"
+
 struct intmap;
 typedef struct intmap intmap;
 
@@ -23,6 +29,7 @@ void *im_find(intmap *, im_key);
 bool im_exists(intmap *, im_key);
 bool im_delete(intmap *, im_key);
 void im_dump_graph(intmap *, const char *);
+/* TODO: Remove dependency on mushtype.h. */
 void im_stats_header(dbref);
 void im_stats(dbref, intmap *, const char *);
-#endif                          /* INTMAP_H */
+#endif /* INTMAP_H */
