@@ -4,20 +4,18 @@
  * \brief Access control lists for PennMUSH.
  * \verbatim
  *
- * The file access.cnf in the game directory will control all 
+ * The file access.cnf in the game directory will control all
  * access-related directives, replacing lockout.cnf and sites.cnf
  *
  * The format of entries in the file will be:
  *
  * wild-host-name    [!]option [!]option [!]option ... # comment
  *
- * A wild-host-name is a wildcard pattern to match hostnames with. 
+ * A wild-host-name is a wildcard pattern to match hostnames with.
  * The wildcard "*" will work like UNIX filename globbing, so
  * *.edu will match all sites with names ending in .edu, and
  * *.*.*.*.* will match all sites with 4 periods in their name.
  * 128.32.*.* will match all sites starting with 128.32 (UC Berkeley).
- * You can also use user@host to match specific users if you know that
- * the host is running ident and you trust its responses (nontrivial).
  *
  * The options that can be specified are:
  * *CONNECT              Allow connections to non-guest players
@@ -50,21 +48,21 @@
  *
  * @sitelock'd sites appear after the line "@sitelock" in the file
  * Using @sitelock writes out the file.
- * 
+ *
  * \endverbatim
  */
 
-#include "config.h"
 #include "copyrite.h"
+#include "access.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
-#ifdef I_SYS_TYPES
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
 #include <fcntl.h>
-#ifdef I_SYS_TIME
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #ifdef TIME_WITH_SYS_TIME
 #include <time.h>
@@ -72,21 +70,21 @@
 #else
 #include <time.h>
 #endif
-#ifdef I_UNISTD
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include "conf.h"
-#include "externs.h"
-#include "mypcre.h"
-#include "access.h"
-#include "mymalloc.h"
-#include "match.h"
-#include "parse.h"
-#include "log.h"
-#include "mushdb.h"
 #include "dbdefs.h"
+#include "externs.h"
 #include "flags.h"
-#include "confmagic.h"
+#include "log.h"
+#include "match.h"
+#include "mushdb.h"
+#include "mymalloc.h"
+#include "mypcre.h"
+#include "notify.h"
+#include "parse.h"
+#include "strutil.h"
 
 /** An access flag. */
 typedef struct a_acsflag acsflag;
@@ -94,86 +92,136 @@ typedef struct a_acsflag acsflag;
  * This structure is used to build a table of access control flags.
  */
 struct a_acsflag {
-  const char *name;             /**< Name of the access flag */
-  bool toggle;                   /**< Is this a negatable flag? */
-  uint32_t flag;                     /**< Bitmask of the flag */
-};
-static acsflag acslist[] = {
-  {"connect", 1, ACS_CONNECT},
-  {"create", 1, ACS_CREATE},
-  {"guest", 1, ACS_GUEST},
-  {"default", 0, ACS_DEFAULT},
-  {"register", 0, ACS_REGISTER},
-  {"suspect", 0, ACS_SUSPECT},
-  {"deny_silent", 0, ACS_DENY_SILENT},
-  {"regexp", 0, ACS_REGEXP},
-  {"god", 1, ACS_GOD},
-  {"wizard", 1, ACS_WIZARD},
-  {"admin", 1, ACS_ADMIN},
-  {NULL, 0, 0}
+  const char *name; /**< Name of the access flag */
+  bool toggle;      /**< Is this a negatable flag? */
+  uint32_t flag;    /**< Bitmask of the flag */
 };
+static acsflag acslist[] = {{"connect", 1, ACS_CONNECT},
+                            {"create", 1, ACS_CREATE},
+                            {"guest", 1, ACS_GUEST},
+                            {"default", 0, ACS_DEFAULT},
+                            {"register", 0, ACS_REGISTER},
+                            {"suspect", 0, ACS_SUSPECT},
+                            {"deny_silent", 0, ACS_DENY_SILENT},
+                            {"regexp", 0, ACS_REGEXP},
+                            {"god", 1, ACS_GOD},
+                            {"wizard", 1, ACS_WIZARD},
+                            {"admin", 1, ACS_ADMIN},
+                            {NULL, 0, 0}};
 
 static struct access *access_top;
 static void free_access_list(void);
 
-extern const unsigned char *tables;
-
 static void
 sitelock_free(struct access *ap)
 {
   mush_free(ap->host, "sitelock.rule.pattern");
-  if (ap->comment)
+  if (ap->comment) {
     mush_free(ap->comment, "sitelock.rule.comment");
-  if (ap->re)
-    free(ap);
+  }
+  if (ap->re) {
+    pcre2_code_free(ap->re);
+  }
+  if (ap->md) {
+    pcre2_match_data_free(ap->md);
+  }
   mush_free(ap, "sitelock.rule");
 }
 
-static struct access *
-sitelock_alloc(const char *host, dbref who,
-               uint32_t can, uint32_t cant,
-               const char *comment, const char **errptr)
-  __attribute_malloc__;
+static struct access *sitelock_alloc(const char *host, dbref who, uint32_t can,
+                                     uint32_t cant, const char *comment,
+                                     const char **errptr) __attribute_malloc__;
 
-    static struct access *sitelock_alloc(const char *host, dbref who,
-                                         uint32_t can, uint32_t cant,
-                                         const char *comment,
-                                         const char **errptr)
+static struct access *
+sitelock_alloc(const char *host, dbref who, uint32_t can, uint32_t cant,
+               const char *comment, const char **errptr)
 {
   struct access *tmp;
+  static PCRE2_UCHAR errstr[120];
+  PCRE2_SIZE erroffset = 0;
+  int errcode;
+
   tmp = mush_malloc(sizeof(struct access), "sitelock.rule");
   if (!tmp) {
-    static const char *memerr = "unable to allocate memory";
-    if (errptr)
+    static const char memerr[] = "unable to allocate memory";
+    if (errptr) {
       *errptr = memerr;
+    }
     return NULL;
   }
+  tmp->re = NULL;
+  tmp->md = NULL;
   tmp->who = who;
   tmp->can = can;
   tmp->cant = cant;
   tmp->host = mush_strdup(host, "sitelock.rule.pattern");
-  if (comment && *comment)
+  if (comment && *comment) {
     tmp->comment = mush_strdup(comment, "sitelock.rule.comment");
-  else
+  } else {
     tmp->comment = NULL;
+  }
   tmp->next = NULL;
 
   if (can & ACS_REGEXP) {
-    int erroffset = 0;
-    tmp->re = pcre_compile(host, 0, errptr, &erroffset, tables);
+    tmp->re =
+      pcre2_compile((const PCRE2_UCHAR *) host, PCRE2_ZERO_TERMINATED,
+                    re_compile_flags, &errcode, &erroffset, re_compile_ctx);
     if (!tmp->re) {
+      if (errptr) {
+        pcre2_get_error_message(errcode, errstr, sizeof errstr);
+        *errptr = (const char *) errstr;
+      }
       sitelock_free(tmp);
       return NULL;
     }
-  } else
-    tmp->re = NULL;
+    pcre2_jit_compile(tmp->re, PCRE2_JIT_COMPLETE);
+    tmp->md = pcre2_match_data_create_from_pattern(tmp->re, NULL);
+  } else {
+    PCRE2_UCHAR *as_re = NULL;
+    PCRE2_SIZE patlen;
+    int r;
+
+    if ((r = pcre2_pattern_convert(
+           (const PCRE2_UCHAR *) tmp->host, PCRE2_ZERO_TERMINATED,
+           PCRE2_CONVERT_GLOB | PCRE2_CONVERT_GLOB_NO_WILD_SEPARATOR, &as_re,
+           &patlen, glob_convert_ctx)) == 0) {
+      /*
+         do_rawlog(LT_ERR, "Compiling '%s' to '%s'", tmp->host,
+         (char *)as_re);
+      */
+      tmp->re = pcre2_compile(as_re, patlen, re_compile_flags | PCRE2_CASELESS,
+                              &errcode, &erroffset, re_compile_ctx);
+      if (tmp->re) {
+        pcre2_jit_compile(tmp->re, PCRE2_JIT_COMPLETE);
+        tmp->md = pcre2_match_data_create_from_pattern(tmp->re, NULL);
+      } else {
+        pcre2_get_error_message(errcode, errstr, sizeof errstr);
+        do_rawlog(LT_ERR,
+                  "Unable to compile sitelock host pattern '%s' as RE '%s': %s",
+                  tmp->host, (char *) as_re, errstr);
+        if (errptr) {
+          *errptr = (const char *) errstr;
+        }
+      }
+      pcre2_converted_pattern_free(as_re);
+    } else {
+      pcre2_get_error_message(r, errstr, sizeof errstr);
+      do_rawlog(LT_ERR,
+                "Unable to compile sitelock host pattern '%s' to regular "
+                "expression: %s",
+                tmp->host, errstr);
+      if (errptr) {
+        *errptr = (const char *) errstr;
+      }
+    }
+  }
 
   return tmp;
 }
 
 static bool
-add_access_node(const char *host, dbref who, uint32_t can,
-                uint32_t cant, const char *comment, const char **errptr)
+add_access_node(const char *host, dbref who, uint32_t can, uint32_t cant,
+                const char *comment, const char **errptr)
 {
   struct access *end, *tmp;
 
@@ -193,7 +241,6 @@ add_access_node(const char *host, dbref who, uint32_t can,
   return true;
 }
 
-
 /** Read the access.cnf file.
  * Initialize the access rules linked list and read in the access.cnf file.
  * \return true if successful, false if not
@@ -206,12 +253,12 @@ read_access_file(void)
   char *p;
   uint32_t can, cant;
   int retval;
-  dbref who;
+  dbref who = AMBIGUOUS;
   char *comment;
   const char *errptr = NULL;
 
   if (access_top) {
-    /* We're reloading the file, so we've got to delete any current 
+    /* We're reloading the file, so we've got to delete any current
      * entries
      */
     free_access_list();
@@ -221,7 +268,7 @@ read_access_file(void)
   release_fd();
   fp = fopen(ACCESS_FILE, FOPEN_READ);
   if (!fp) {
-    do_rawlog(LT_ERR, T("Access file %s not found."), ACCESS_FILE);
+    do_rawlog(LT_ERR, "Access file %s not found.", ACCESS_FILE);
     retval = 0;
   } else {
     do_rawlog(LT_ERR, "Reading %s", ACCESS_FILE);
@@ -233,7 +280,7 @@ read_access_file(void)
         *p = '\0';
       /* Find beginning of line; ignore blank lines */
       p = buf;
-      if (*p && isspace((unsigned char) *p))
+      if (*p && isspace(*p))
         p++;
       if (*p && *p != '#') {
         can = cant = 0;
@@ -242,16 +289,15 @@ read_access_file(void)
         if (!strncasecmp(p, "@sitelock", 9)) {
           if (!add_access_node("@sitelock", AMBIGUOUS, ACS_SITELOCK, 0, "",
                                &errptr))
-            do_log(LT_ERR, GOD, GOD, T("Failed to add sitelock node: %s"),
-                   errptr);
+            do_log(LT_ERR, GOD, GOD, "Failed to add sitelock node: %s", errptr);
         } else {
           if ((comment = strchr(p, '#'))) {
             *comment++ = '\0';
-            while (*comment && isspace((unsigned char) *comment))
+            while (*comment && isspace(*comment))
               comment++;
           }
           /* Move past the host name */
-          while (*p && !isspace((unsigned char) *p))
+          while (*p && !isspace(*p))
             p++;
           if (*p)
             *p++ = '\0';
@@ -259,8 +305,7 @@ read_access_file(void)
             /* Nothing listed, so assume we can't do anything! */
             cant = ACS_DEFAULT;
           if (!add_access_node(buf, who, can, cant, comment, &errptr))
-            do_log(LT_ERR, GOD, GOD, T("Failed to add access node: %s"),
-                   errptr);
+            do_log(LT_ERR, GOD, GOD, "Failed to add access node: %s", errptr);
         }
       }
     }
@@ -287,7 +332,7 @@ write_access_file(void)
   release_fd();
   fp = fopen(tmpf, FOPEN_WRITE);
   if (!fp) {
-    do_log(LT_ERR, GOD, GOD, T("Unable to open %s."), tmpf);
+    do_log(LT_ERR, GOD, GOD, "Unable to open %s.", tmpf);
   } else {
     for (ap = access_top; ap; ap = ap->next) {
       if (strcmp(ap->host, "@sitelock") == 0) {
@@ -329,22 +374,8 @@ write_access_file(void)
   return;
 }
 
-#ifdef FORCE_IPV4
-static char *
-ip4_to_ip6(const char *addr)
-{
-  static char tbuf1[BUFFER_LEN];
-  char *bp;
-  bp = tbuf1;
-  safe_format(tbuf1, &bp, "::ffff:%s", addr);
-  *bp = '\0';
-  return tbuf1;
-}
-#endif
-
-
 /** Decide if a host can access someway.
- * \param hname a host or user+host pattern.
+ * \param hname a host.
  * \param flag the access type we're testing.
  * \param who the player attempting access.
  * \retval 1 access permitted.
@@ -353,8 +384,6 @@ ip4_to_ip6(const char *addr)
  * Given a hostname and a flag decide if the host can do it.
  * Here's how it works:
  * We run the linked list and take the first match.
- *  (If the hostname is user@host, we try to match both user@host
- *   and just host to each line in the file.)
  * If we make a match, and the line tells us whether the site can/can't
  *   do the action, we're done.
  * Otherwise, we assume that the host can do any toggleable option
@@ -367,26 +396,22 @@ site_can_access(const char *hname, uint32_t flag, dbref who)
 {
   struct access *ap;
   acsflag *c;
-  const char *p;
+  size_t hlen;
 
   if (!hname || !*hname)
     return 0;
 
-  if ((p = strchr(hname, '@')))
-    p++;
+  hlen = strlen(hname);
 
   for (ap = access_top; ap; ap = ap->next) {
     if (ap->can & ACS_SITELOCK)
       continue;
-    if ((ap->can & ACS_REGEXP)
-        ? (qcomp_regexp_match(ap->re, hname)
-           || (p && qcomp_regexp_match(ap->re, p)))
-        : (quick_wild(ap->host, hname)
-           || (p && quick_wild(ap->host, p)))
-        && (ap->who == AMBIGUOUS || ap->who == who)) {
+    if ((ap->re ? qcomp_regexp_match(ap->re, ap->md, hname, hlen)
+                : quick_wild(ap->host, hname)) &&
+        (ap->who == AMBIGUOUS || ap->who == who)) {
       /* Got one */
       if (flag & ACS_CONNECT) {
-        if ((ap->cant & ACS_GOD) && God(who))   /* God can't connect from here */
+        if ((ap->cant & ACS_GOD) && God(who)) /* God can't connect from here */
           return 0;
         else if ((ap->cant & ACS_WIZARD) && Wizard(who))
           /* Wiz can't connect from here */
@@ -415,7 +440,6 @@ site_can_access(const char *hname, uint32_t flag, dbref who)
   return 1;
 }
 
-
 /** Return the first access rule that matches a host.
  * \param hname a host or user+host pattern.
  * \param who the player attempting access.
@@ -426,25 +450,21 @@ struct access *
 site_check_access(const char *hname, dbref who, int *rulenum)
 {
   struct access *ap;
-  const char *p;
+  size_t hlen;
 
   *rulenum = 0;
   if (!hname || !*hname)
     return 0;
 
-  if ((p = strchr(hname, '@')))
-    p++;
+  hlen = strlen(hname);
 
   for (ap = access_top; ap; ap = ap->next) {
     (*rulenum)++;
     if (ap->can & ACS_SITELOCK)
       continue;
-    if (((ap->can & ACS_REGEXP)
-         ? (qcomp_regexp_match(ap->re, hname)
-            || (p && qcomp_regexp_match(ap->re, p)))
-         : (quick_wild(ap->host, hname)
-            || (p && quick_wild(ap->host, p))))
-        && (ap->who == AMBIGUOUS || ap->who == who)) {
+    if ((ap->re ? qcomp_regexp_match(ap->re, ap->md, hname, hlen)
+                : quick_wild(ap->host, hname)) &&
+        (ap->who == AMBIGUOUS || ap->who == who)) {
       /* Got one */
       return ap;
     }
@@ -463,7 +483,7 @@ site_check_access(const char *hname, dbref who, int *rulenum)
  */
 void
 format_access(struct access *ap, int rulenum,
-              dbref who __attribute__ ((__unused__)), char *buff, char **bp)
+              dbref who __attribute__((__unused__)), char *buff, char **bp)
 {
   if (ap) {
     safe_format(buff, bp, T("Matched line %d: %s %s"), rulenum, ap->host,
@@ -504,7 +524,6 @@ format_access(struct access *ap, int rulenum,
   }
 }
 
-
 /** Add an access rule to the linked list.
  * \param player enactor.
  * \param host host pattern to add.
@@ -528,7 +547,6 @@ add_access_sitelock(dbref player, const char *host, dbref who, uint32_t can,
   struct access *tmp;
   const char *errptr = NULL;
 
-
   tmp = sitelock_alloc(host, who, can, cant, "", &errptr);
 
   if (!tmp) {
@@ -538,7 +556,8 @@ add_access_sitelock(dbref player, const char *host, dbref who, uint32_t can,
 
   if (!access_top) {
     /* Add to the beginning, but first add a sitelock marker */
-    if (!add_access_node("@sitelock", AMBIGUOUS, ACS_SITELOCK, 0, "", &errptr)) {
+    if (!add_access_node("@sitelock", AMBIGUOUS, ACS_SITELOCK, 0, "",
+                         &errptr)) {
       notify_format(player, T("Unable to add @sitelock separator: %s"), errptr);
       return 0;
     }
@@ -597,7 +616,7 @@ remove_access_sitelock(const char *pattern)
     rulenum++;
     next = ap->next;
     if (deletethis == -1 ? (strcasecmp(pattern, ap->host) == 0)
-        : deletethis == rulenum) {
+                         : deletethis == rulenum) {
       n++;
       sitelock_free(ap);
       if (prev)
@@ -629,7 +648,6 @@ free_access_list(void)
   access_top = NULL;
 }
 
-
 /** Display the access list.
  * \param player enactor.
  * Sends the complete access list to the player.
@@ -661,16 +679,14 @@ do_list_access(dbref player)
         }
       }
       *bp = '\0';
-      notify_format(player,
-                    "%3d SITE: %-20s  DBREF: %-6s FLAGS:%s", rulenum,
+      notify_format(player, T("%3d SITE: %-20s  DBREF: %-6s FLAGS:%s"), rulenum,
                     ap->host, unparse_dbref(ap->who), flaglist);
-      notify_format(player, " COMMENT: %s", ap->comment ? ap->comment : "");
+      notify_format(player, T(" COMMENT: %s"), ap->comment ? ap->comment : "");
     } else {
-      notify(player,
-             T
-             ("---- @sitelock will add sites immediately below this line ----"));
+      notify(
+        player,
+        T("---- @sitelock will add sites immediately below this line ----"));
     }
-
   }
   if (rulenum == 0) {
     notify(player, T("There are no access rules."));
@@ -709,9 +725,9 @@ parse_access_options(const char *opts, dbref *who, uint32_t *can,
   while ((w = split_token(&p, ' '))) {
     found = 0;
 
-    if (first && who) {         /* Check for a character */
+    if (first && who) { /* Check for a character */
       first = 0;
-      if (is_integer(w)) {      /* We have a dbref */
+      if (is_strict_integer(w)) { /* We have a dbref */
         *who = parse_integer(w);
         if (*who != AMBIGUOUS && !GoodObject(*who))
           *who = AMBIGUOUS;
@@ -747,7 +763,7 @@ parse_access_options(const char *opts, dbref *who, uint32_t *can,
       if (GoodObject(player))
         notify_format(player, T("Unknown access option: %s"), w);
       else
-        do_log(LT_ERR, GOD, GOD, T("Unknown access flag: %s"), w);
+        do_log(LT_ERR, GOD, GOD, "Unknown access flag: %s", w);
     } else {
       totalfound += found;
     }
