@@ -7,17 +7,17 @@
  */
 
 #include "copyrite.h"
-#include "config.h"
+#include "bufferq.h"
 
 #include <stdio.h>
-#ifdef I_UNISTD
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <string.h>
 #include <stddef.h>
 #include <stdarg.h>
 #include <stdlib.h>
-#ifdef I_SYS_TIME
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #ifdef TIME_WITH_SYS_TIME
 #include <time.h>
@@ -25,20 +25,18 @@
 #else
 #include <time.h>
 #endif
-#ifdef I_SYS_TYPES
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
 
 #include "conf.h"
+#include "dbdefs.h"
 #include "externs.h"
 #include "flags.h"
-#include "dbdefs.h"
-#include "bufferq.h"
-#include "mymalloc.h"
 #include "log.h"
-#include "confmagic.h"
+#include "mymalloc.h"
 
-#define BUFFERQLINEOVERHEAD     (2*sizeof(int)+sizeof(time_t)+sizeof(dbref))
+#define BUFFERQLINEOVERHEAD (2 * sizeof(int) + sizeof(time_t) + sizeof(dbref))
 
 static void shift_bufferq(BUFFERQ *bq, int space_needed);
 
@@ -75,7 +73,6 @@ add_to_bufferq(BUFFERQ *bq, int type, dbref player, const char *msg)
   bq->num_buffered++;
 }
 
-
 static void
 shift_bufferq(BUFFERQ *bq, int space_needed)
 {
@@ -113,7 +110,6 @@ shift_bufferq(BUFFERQ *bq, int space_needed)
 }
 
 /** Allocate memory for a buffer queue to hold a given number of lines.
- * \param bq pointer to buffer queue.
  * \param lines lines to allocate for buffer queue.
  * \retval address of allocated buffer queue.
  */
@@ -179,8 +175,6 @@ reallocate_bufferq(BUFFERQ *bq, int lines)
   return bq;
 }
 
-
-
 /** Iterate through messages in a bufferq.
  * This function returns the next message in a bufferq, given
  * a pointer to the start of entry (which is modified).
@@ -196,8 +190,7 @@ reallocate_bufferq(BUFFERQ *bq, int lines)
  * \return next message text or NULL if no more.
  */
 char *
-iter_bufferq(BUFFERQ *bq, char **p, dbref *player, int *type,
-             time_t * timestamp)
+iter_bufferq(BUFFERQ *bq, char **p, dbref *player, int *type, time_t *timestamp)
 {
   static char tbuf1[BUFFER_LEN];
   int size;
@@ -209,7 +202,7 @@ iter_bufferq(BUFFERQ *bq, char **p, dbref *player, int *type,
     return NULL;
 
   if (!*p)
-    *p = bq->buffer;            /* Reset to beginning */
+    *p = bq->buffer; /* Reset to beginning */
 
   memcpy(&size, *p, sizeof(size));
   *p += sizeof(size);
@@ -242,7 +235,7 @@ bufferq_blocks(BUFFERQ *bq)
  * \return line count
  */
 int
-bufferq_lines(BUFFERQ *bp)
+bufferq_lines(BUFFERQ *bq)
 {
   int lines = 0;
   char *p = NULL;
@@ -250,11 +243,10 @@ bufferq_lines(BUFFERQ *bp)
   int type;
   time_t t;
 
-
-  if (isempty_bufferq(bp))
+  if (isempty_bufferq(bq))
     return 0;
 
-  while (iter_bufferq(bp, &p, &player, &type, &t))
+  while (iter_bufferq(bq, &p, &player, &type, &t))
     lines++;
 
   return lines;
