@@ -5,46 +5,75 @@
  *
  *
  */
+
 #include "copyrite.h"
 
-#include "config.h"
-#include <time.h>
-#include <string.h>
 #include <ctype.h>
-#include "conf.h"
-#include "case.h"
-#include "externs.h"
-#include "version.h"
-#include "extchat.h"
-#include "htab.h"
-#include "flags.h"
-#include "dbdefs.h"
-#include "parse.h"
-#include "function.h"
-#include "command.h"
-#include "game.h"
-#include "attrib.h"
-#include "ansi.h"
-#include "match.h"
-#ifdef HAS_OPENSSL
-#include <openssl/sha.h>
-#include <openssl/evp.h>
-#include <openssl/bio.h>
+#ifdef WIN32
+#include <Windows.h>
+#include <Wincrypt.h>
+#include <Bcrypt.h>
 #else
-#include "shs.h"
+#include <openssl/bio.h>
+#include <openssl/sha.h>
 #endif
-#include "confmagic.h"
+#include <openssl/evp.h>
+#include <openssl/hmac.h>
+#include <string.h>
+#include <time.h>
 
+#include "ansi.h"
+#include "attrib.h"
+#include "case.h"
+#include "command.h"
+#include "conf.h"
+#include "dbdefs.h"
+#include "extchat.h"
+#include "externs.h"
+#include "flags.h"
+#include "function.h"
+#include "game.h"
+#include "htab.h"
+#include "match.h"
+#include "mymalloc.h"
+#include "parse.h"
+#include "sort.h"
+#include "strutil.h"
+#include "charclass.h"
 
-static char *crunch_code(char *code);
-static char *crypt_code(char *code, char *text, int type);
-static void safe_sha0(const char *text, size_t len, char *buff, char **bp);
-static void safe_hexchar(unsigned char c, char *buff, char **bp);
+char *crunch_code(char *code);
+char *crypt_code(char *code, char *text, int type);
+bool decode_base64(char *encoded, int len, bool printonly, char *buff,
+                   char **bp);
 
 static bool
 encode_base64(const char *input, int len, char *buff, char **bp)
 {
-#ifdef HAVE_SSL
+#ifdef WIN32
+  LPTSTR encoded;
+  DWORD encodedlen = 0;
+
+  if (!CryptBinaryToString((const BYTE *) input, len,
+                           CRYPT_STRING_BASE64 | CRYPT_STRING_NOCRLF, NULL,
+                           &encodedlen)) {
+    safe_str(T("#-1 ENCODING ERROR"), buff, bp);
+    return false;
+  }
+
+  encoded = mush_malloc(encodedlen, "string");
+  if (!CryptBinaryToString((const BYTE *) input, len,
+                           CRYPT_STRING_BASE64 | CRYPT_STRING_NOCRLF, encoded,
+                           &encodedlen)) {
+    mush_free(encoded, "string");
+    safe_str(T("#-1 ENCODING ERROR"), buff, bp);
+    return false;
+  }
+
+  safe_strl(encoded, encodedlen, buff, bp);
+  mush_free(encoded, "string");
+
+  return true;
+#else
   BIO *bio, *b64, *bmem;
   char *membuf;
 
@@ -71,7 +100,7 @@ encode_base64(const char *input, int len, char *buff, char **bp)
     return false;
   }
 
-  len = BIO_flush(bio);
+  (void) BIO_flush(bio);
 
   len = BIO_get_mem_data(bmem, &membuf);
 
@@ -80,18 +109,64 @@ encode_base64(const char *input, int len, char *buff, char **bp)
   BIO_free_all(bio);
 
   return true;
-#else
-  safe_str(T("#-1 FUNCTION DISABLED"), buff, bp);
-  return false;
 #endif
 }
 
 extern char valid_ansi_codes[UCHAR_MAX + 1];
 
-static bool
-decode_base64(char *encoded, int len, char *buff, char **bp)
+bool
+decode_base64(char *encoded, int len, bool printonly, char *buff, char **bp)
 {
-#ifdef HAVE_SSL
+#ifdef WIN32
+  BYTE *decoded;
+  DWORD dlen = 0;
+  char *sbp = *bp;
+
+  if (!CryptStringToBinary((LPCTSTR) encoded, len, CRYPT_STRING_BASE64, NULL,
+                           &dlen, NULL, NULL)) {
+    safe_str(T("#-1 DECODING ERROR"), buff, bp);
+    return false;
+  }
+
+  decoded = mush_malloc(dlen + 1, "string");
+  if (!CryptStringToBinary((LPCTSTR) encoded, len, CRYPT_STRING_BASE64, decoded,
+                           &dlen, NULL, NULL)) {
+    mush_free(decoded, "string");
+    safe_str(T("#-1 DECODING ERROR"), buff, bp);
+    return false;
+  }
+  decoded[dlen] = '\0';
+
+  for (DWORD n = 0; n < dlen; n++) {
+    if (decoded[n] == TAG_START) {
+      DWORD end;
+      n += 1;
+      for (end = n; end < dlen; end++) {
+        if (decoded[end] == TAG_END)
+          break;
+      }
+      if (end == dlen || decoded[n] != MARKUP_COLOR) {
+        mush_free(decoded, "string");
+        *bp = sbp;
+        safe_str(T("#-1 CONVERSION ERROR"), buff, bp);
+        return false;
+      }
+      for (; n < end; n++) {
+        if (!valid_ansi_codes[decoded[n]]) {
+          mush_free(decoded, "string");
+          *bp = sbp;
+          safe_str(T("#-1 CONVERSION ERROR"), buff, bp);
+          return false;
+        }
+      }
+      n = end;
+    } else if (printonly && !char_isprint(decoded[n]))
+      decoded[n] = '?';
+  }
+  safe_strl((const char *) decoded, dlen, buff, bp);
+  mush_free(decoded, "string");
+  return true;
+#else
   BIO *bio, *b64, *bmem;
   char *sbp;
 
@@ -108,7 +183,8 @@ decode_base64(char *encoded, int len, char *buff, char **bp)
     BIO_free(b64);
     return false;
   }
-  /*  len = BIO_set_close(bmem, BIO_NOCLOSE); This makes valgrind report a memory leak. */
+  /*  len = BIO_set_close(bmem, BIO_NOCLOSE); This makes valgrind report a
+   * memory leak. */
 
   bio = BIO_push(b64, bmem);
 
@@ -135,7 +211,7 @@ decode_base64(char *encoded, int len, char *buff, char **bp)
             return false;
           }
           for (; n < end; n++) {
-            if (!valid_ansi_codes[(unsigned char) decoded[n]]) {
+            if (!valid_ansi_codes[decoded[n]]) {
               BIO_free_all(bio);
               *bp = sbp;
               safe_str(T("#-1 CONVERSION ERROR"), buff, bp);
@@ -143,7 +219,7 @@ decode_base64(char *encoded, int len, char *buff, char **bp)
             }
           }
           n = end;
-        } else if (!isprint((unsigned char) decoded[n]))
+        } else if (printonly && !char_isprint(decoded[n]))
           decoded[n] = '?';
       }
       safe_strl(decoded, dlen, buff, bp);
@@ -160,26 +236,17 @@ decode_base64(char *encoded, int len, char *buff, char **bp)
   BIO_free_all(bio);
 
   return true;
-#else
-  safe_str(T("#-1 FUNCTION DISABLED"), buff, bp);
-  return false;
 #endif
 }
 
 /* Encode a string in base64 */
-FUNCTION(fun_encode64)
-{
-  encode_base64(args[0], arglens[0], buff, bp);
-}
+FUNCTION(fun_encode64) { encode_base64(args[0], arglens[0], buff, bp); }
 
 /* Decode a string from base64 */
-FUNCTION(fun_decode64)
-{
-  decode_base64(args[0], arglens[0], buff, bp);
-}
+FUNCTION(fun_decode64) { decode_base64(args[0], arglens[0], 1, buff, bp); }
 
 /* Copy over only alphanumeric chars */
-static char *
+char *
 crunch_code(char *code)
 {
   char *in;
@@ -188,12 +255,7 @@ crunch_code(char *code)
 
   out = output;
   in = code;
-  while (*in) {
-    while (*in == ESC_CHAR) {
-      while (*in && *in != 'm')
-        in++;
-      in++;                     /* skip 'm' */
-    }
+  WALK_ANSI_STRING (in) {
     if ((*in >= 32) && (*in <= 126)) {
       *out++ = *in;
     }
@@ -203,7 +265,7 @@ crunch_code(char *code)
   return output;
 }
 
-static char *
+char *
 crypt_code(char *code, char *text, int type)
 {
   static char textbuff[BUFFER_LEN];
@@ -213,7 +275,7 @@ crypt_code(char *code, char *text, int type)
   int mod = end - start + 1;
   char *p, *q, *r;
 
-  if (!text && !*text)
+  if (!(text && *text))
     return (char *) "";
   if (!code || !*code)
     return text;
@@ -244,29 +306,6 @@ crypt_code(char *code, char *text, int type)
   return textbuff;
 }
 
-static void
-safe_sha0(const char *text, size_t len, char *buff, char **bp)
-{
-#ifdef HAS_OPENSSL
-  unsigned char hash[SHA_DIGEST_LENGTH];
-  int n;
-
-  SHA((unsigned char *) text, len, hash);
-
-  for (n = 0; n < SHA_DIGEST_LENGTH; n++)
-    safe_hexchar(hash[n], buff, bp);
-#else
-  SHS_INFO shsInfo;
-  shsInfo.reverse_wanted = (BYTE) options.reverse_shs;
-  shsInit(&shsInfo);
-  shsUpdate(&shsInfo, (const BYTE *) text, len);
-  shsFinal(&shsInfo);
-  safe_format(buff, bp, "%0x%0x%0x%0x%0x", shsInfo.digest[0],
-              shsInfo.digest[1], shsInfo.digest[2], shsInfo.digest[3],
-              shsInfo.digest[4]);
-#endif
-}
-
 FUNCTION(fun_encrypt)
 {
   char tbuff[BUFFER_LEN], *tp;
@@ -304,7 +343,7 @@ FUNCTION(fun_decrypt)
 
   if (nargs == 3 && parse_boolean(args[2])) {
     tp = tbuff;
-    if (!decode_base64(args[0], arglens[0], tbuff, &tp)) {
+    if (!decode_base64(args[0], arglens[0], 1, tbuff, &tp)) {
       safe_strl(tbuff, tp - tbuff, buff, bp);
       return;
     }
@@ -337,49 +376,107 @@ FUNCTION(fun_checkpass)
 
 FUNCTION(fun_sha0)
 {
-  safe_sha0(args[0], arglens[0], buff, bp);
-}
+#ifdef HAVE_SHA
+  uint8_t hash[SHA_DIGEST_LENGTH];
 
-FUNCTION(fun_digest)
-{
-#ifdef HAS_OPENSSL
-  EVP_MD_CTX ctx;
-  const EVP_MD *mp;
-  unsigned char md[EVP_MAX_MD_SIZE];
-  unsigned int n, len = 0;
-
-  if ((mp = EVP_get_digestbyname(args[0])) == NULL) {
-    safe_str(T("#-1 UNSUPPORTED DIGEST TYPE"), buff, bp);
-    return;
-  }
-
-  EVP_DigestInit(&ctx, mp);
-  EVP_DigestUpdate(&ctx, args[1], arglens[1]);
-  EVP_DigestFinal(&ctx, md, &len);
-
-  for (n = 0; n < len; n++) {
-    safe_hexchar(md[n], buff, bp);
-  }
+  SHA((uint8_t *) args[0], arglens[0], hash);
 
+  safe_hexstr(hash, SHA_DIGEST_LENGTH, buff, bp);
 #else
-  if (strcmp(args[0], "sha") == 0) {
-    safe_sha0(args[1], arglens[1], buff, bp);
-  } else {
-    safe_str(T("#-1 UNSUPPORTED DIGEST TYPE"), buff, bp);
-  }
+  safe_str(T("#-1 NOT SUPPORTED"), buff, bp);
 #endif
 }
 
+/* From mycrypt.c */
+int safe_hash_byname(const char *algo, const char *plaintext, int len,
+                     char *buff, char **bp, bool inplace_err);
+
+#ifdef HAVE_EVP_MD_DO_ALL
+#define CAN_LIST_DIGESTS
+
+#ifndef WIN32
 static void
-safe_hexchar(unsigned char c, char *buff, char **bp)
+list_dgst_populate(const EVP_MD *m,
+                   const char *from __attribute__((__unused__)),
+                   const char *to __attribute__((__unused__)), void *data)
 {
-  const char *digits = "0123456789abcdef";
-  if (*bp - buff < BUFFER_LEN - 1) {
-    **bp = digits[c >> 4];
-    (*bp)++;
+  HASHTAB *digests = data;
+  if (m)
+    hash_add(digests, EVP_MD_name(m), "foo");
+}
+#endif
+
+#endif
+
+FUNCTION(fun_digest)
+{
+  if (nargs == 1 && strcmp(args[0], "list") == 0) {
+#ifdef WIN32
+    safe_str("MD2 MD4 MD5 SHA1 SHA256 SHA384 SHA512", buff, bp);
+    return;
+#elif defined(CAN_LIST_DIGESTS)
+    HASHTAB digests_tab;
+    const char **digests;
+    const char *d;
+    int m, n;
+
+    hashinit(&digests_tab, 100);
+    EVP_MD_do_all(list_dgst_populate, &digests_tab);
+    digests = mush_calloc(digests_tab.entries, sizeof(char *), "digest.list");
+
+    for (n = 0, d = hash_firstentry_key(&digests_tab); d;
+         n += 1, d = hash_nextentry_key(&digests_tab))
+      digests[n] = d;
+
+    qsort(digests, n, sizeof(char *), stri_comp);
+
+    for (m = 0; m < n; m += 1) {
+      if (m > 0)
+        safe_chr(' ', buff, bp);
+      safe_str(digests[m], buff, bp);
+    }
+
+    mush_free(digests, "digest.list");
+    hashfree(&digests_tab);
+#else
+    safe_str(T("#-1 LISTING NOT SUPPORTED"), buff, bp);
+#endif
+  } else if (nargs == 2)
+    safe_hash_byname(args[0], args[1], arglens[1], buff, bp, 1);
+  else
+    safe_str(T("#-1 INVALID ARGUMENT"), buff, bp);
+}
+
+FUNCTION(fun_hmac)
+{
+  unsigned char md[EVP_MAX_MD_SIZE];
+  unsigned int md_len = EVP_MAX_MD_SIZE;
+  const EVP_MD *hash;
+  bool base16 = 1;
+
+  if (nargs == 4) {
+    if (sqlite3_stricmp(args[3], "base16") == 0) {
+      base16 = 1;
+    } else if (sqlite3_stricmp(args[3], "base64") == 0) {
+      base16 = 0;
+    } else {
+      safe_str(T("#-1 INVALID ARGUMENT"), buff, bp);
+      return;
+    }
   }
-  if (*bp - buff < BUFFER_LEN - 1) {
-    **bp = digits[c & 0x0F];
-    (*bp)++;
+
+  /* TODO: Windows BCrypt version */
+  hash = EVP_get_digestbyname(args[0]);
+  if (!hash) {
+    safe_str(T("#-1 UNSUPPORTED DIGEST TYPE"), buff, bp);
+    return;
+  }
+  HMAC(hash, args[1], arglens[1], (unsigned char *) args[2], arglens[2], md,
+       &md_len);
+
+  if (base16) {
+    safe_hexstr(md, md_len, buff, bp);
+  } else {
+    encode_base64((const char *) md, md_len, buff, bp);
   }
 }
