@@ -28,118 +28,86 @@
  */
 
 #include "copyrite.h"
-#include "config.h"
+#include "lock.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include "conf.h"
-#include "externs.h"
-#include "boolexp.h"
-#include "mushdb.h"
 #include "attrib.h"
+#include "boolexp.h"
+#include "conf.h"
 #include "dbdefs.h"
-#include "lock.h"
-#include "match.h"
-#include "log.h"
+#include "externs.h"
 #include "flags.h"
-#include "mymalloc.h"
-#include "strtree.h"
-#include "privtab.h"
-#include "parse.h"
 #include "htab.h"
-#include "confmagic.h"
-
+#include "log.h"
+#include "match.h"
+#include "mushdb.h"
+#include "mymalloc.h"
+#include "notify.h"
+#include "parse.h"
+#include "privtab.h"
+#include "strtree.h"
+#include "strutil.h"
 
 /* If any lock_type ever contains the character '|', reading in locks
  * from the db will break.
  */
-lock_type Basic_Lock = "Basic";     /**< Name of basic lock */
-lock_type Enter_Lock = "Enter";     /**< Name of enter lock */
-lock_type Use_Lock = "Use";         /**< Name of use lock */
-lock_type Zone_Lock = "Zone";       /**< Name of zone lock */
-lock_type Page_Lock = "Page";       /**< Name of page lock */
-lock_type Tport_Lock = "Teleport";  /**< Name of teleport lock */
-lock_type Speech_Lock = "Speech";   /**< Name of speech lock */
-lock_type Listen_Lock = "Listen";   /**< Name of listen lock */
-lock_type Command_Lock = "Command"; /**< Name of command lock */
-lock_type Parent_Lock = "Parent";   /**< Name of parent lock */
-lock_type Link_Lock = "Link";       /**< Name of link lock */
-lock_type Leave_Lock = "Leave";     /**< Name of leave lock */
-lock_type Drop_Lock = "Drop";       /**< Name of drop lock */
-lock_type Give_Lock = "Give";       /**< Name of give lock */
-lock_type Mail_Lock = "Mail";       /**< Name of mail lock */
-lock_type Follow_Lock = "Follow";   /**< Name of follow lock */
-lock_type Examine_Lock = "Examine"; /**< Name of examine lock */
-lock_type Chzone_Lock = "Chzone";   /**< Name of chzone lock */
-lock_type Forward_Lock = "Forward"; /**< Name of forward lock */
-lock_type Control_Lock = "Control"; /**< Name of control lock */
-lock_type Dropto_Lock = "Dropto";   /**< Name of dropto lock */
-lock_type Destroy_Lock = "Destroy"; /**< Name of destroy lock */
-lock_type Interact_Lock = "Interact"; /**< Name of interaction lock */
+lock_type Basic_Lock = "Basic";             /**< Name of basic lock */
+lock_type Enter_Lock = "Enter";             /**< Name of enter lock */
+lock_type Use_Lock = "Use";                 /**< Name of use lock */
+lock_type Zone_Lock = "Zone";               /**< Name of zone lock */
+lock_type Page_Lock = "Page";               /**< Name of page lock */
+lock_type Tport_Lock = "Teleport";          /**< Name of teleport lock */
+lock_type Speech_Lock = "Speech";           /**< Name of speech lock */
+lock_type Listen_Lock = "Listen";           /**< Name of listen lock */
+lock_type Command_Lock = "Command";         /**< Name of command lock */
+lock_type Parent_Lock = "Parent";           /**< Name of parent lock */
+lock_type Link_Lock = "Link";               /**< Name of link lock */
+lock_type Leave_Lock = "Leave";             /**< Name of leave lock */
+lock_type Drop_Lock = "Drop";               /**< Name of drop lock */
+lock_type Give_Lock = "Give";               /**< Name of give lock */
+lock_type From_Lock = "From";               /**< Name of from lock */
+lock_type Pay_Lock = "Pay";                 /**< Name of pay lock */
+lock_type Receive_Lock = "Receive";         /**< Name of receive lock */
+lock_type Mail_Lock = "Mail";               /**< Name of mail lock */
+lock_type Follow_Lock = "Follow";           /**< Name of follow lock */
+lock_type Examine_Lock = "Examine";         /**< Name of examine lock */
+lock_type Chzone_Lock = "Chzone";           /**< Name of chzone lock */
+lock_type Forward_Lock = "Forward";         /**< Name of forward lock */
+lock_type Control_Lock = "Control";         /**< Name of control lock */
+lock_type Dropto_Lock = "Dropto";           /**< Name of dropto lock */
+lock_type Destroy_Lock = "Destroy";         /**< Name of destroy lock */
+lock_type Interact_Lock = "Interact";       /**< Name of interaction lock */
 lock_type MailForward_Lock = "MailForward"; /**< Name of mailforward lock */
-lock_type Take_Lock = "Take"; /**< Name of take lock */
+lock_type Take_Lock = "Take";               /**< Name of take lock */
+lock_type Open_Lock = "Open";               /**< Name of open lock */
+lock_type Filter_Lock = "Filter";           /**< Name of filter lock */
+lock_type InFilter_Lock = "InFilter";       /**< Name of infilter lock */
+lock_type DropIn_Lock = "DropIn";           /**< Name of the DropIn lock */
+lock_type Chown_Lock = "Chown";             /**< Name of Chown lock */
 
- /** Table of lock names and permissions */
-lock_list lock_types[] = {
-  {"Basic", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Enter", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Use", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Zone", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Page", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Teleport", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Speech", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Listen", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Command", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Parent", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Link", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Leave", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Drop", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Give", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Mail", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Follow", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Examine", TRUE_BOOLEXP, GOD, LF_PRIVATE | LF_OWNER, NULL},
-  {"Chzone", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Forward", TRUE_BOOLEXP, GOD, LF_PRIVATE | LF_OWNER, NULL},
-  {"Control", TRUE_BOOLEXP, GOD, LF_PRIVATE | LF_OWNER, NULL},
-  {"Dropto", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Destroy", TRUE_BOOLEXP, GOD, LF_PRIVATE | LF_OWNER, NULL},
-  {"Interact", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"MailForward", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {"Take", TRUE_BOOLEXP, GOD, LF_PRIVATE, NULL},
-  {NULL, TRUE_BOOLEXP, GOD, 0, NULL}
-};
+#include "lock_tab.h"
 
 HASHTAB htab_locks;
 
- /** Table of base attributes associated with success and failure of
-  * locks. These are the historical ones; we automatically generate
+/**
+ * \verbatim
+ * Table of base attributes associated with success and failure of
+ * locks. These are the historical ones; we automatically generate
  * such attribute names for those that aren't in this table using
  * <lock>_LOCK`<message>
+ * \endverbatim
  */
-const LOCKMSGINFO lock_msgs[] = {
-  {"Basic", "SUCCESS", "FAILURE"},
-  {"Enter", "ENTER", "EFAIL"},
-  {"Use", "USE", "UFAIL"},
-  {"Leave", "LEAVE", "LFAIL"},
-  {NULL, NULL, NULL}
-};
+const LOCKMSGINFO lock_msgs[] = {{"Basic", "SUCCESS", "FAILURE"},
+                                 {"Enter", "ENTER", "EFAIL"},
+                                 {"Use", "USE", "UFAIL"},
+                                 {"Leave", "LEAVE", "LFAIL"},
+                                 {NULL, NULL, NULL}};
 
-/** Table of lock permissions */
-PRIV lock_privs[] = {
-  {"visual", 'v', LF_VISUAL, LF_VISUAL},
-  {"no_inherit", 'i', LF_PRIVATE, LF_PRIVATE},
-  {"no_clone", 'c', LF_NOCLONE, LF_NOCLONE},
-  {"wizard", 'w', LF_WIZARD, LF_WIZARD},
-  /*  {"owner", 'o', LF_OWNER, LF_OWNER}, */
-  {"locked", '+', LF_LOCKED, LF_LOCKED},
-  {NULL, '\0', 0, 0}
-};
-
-StrTree lock_names;  /**< String tree of lock names */
+StrTree lock_names; /**< String tree of lock names */
 
 static void free_one_lock_list(lock_list *ll);
-static lock_type check_lock_type(dbref player, dbref thing, lock_type name);
 static int delete_lock(dbref player, dbref thing, lock_type type);
 static int can_write_lock(dbref player, dbref thing, lock_list *lock);
 static lock_list *getlockstruct_noparent(dbref thing, lock_type type);
@@ -150,34 +118,47 @@ static void free_lock(lock_list *ll);
 
 extern int unparsing_boolexp;
 
+static int
+lock_compare(const void *a, const void *b)
+{
+  const lock_list *la = *(lock_list *const *) a, *lb = *(lock_list *const *) b;
+  return strcmp(la->type, lb->type);
+}
+
 /** Return a list of all available locks
  * \param buff the buffer
  * \param bp a pointer to the current position in the buffer.
+ * \param name if not NULL, only show locks with this prefix
  */
 void
 list_locks(char *buff, char **bp, const char *name)
 {
-  char rbuff[BUFFER_LEN];
-  char *rp;
-  int first = 1;
-  const lock_list *ptr;
-  rp = rbuff;
-  for (ptr = hash_firstentry(&htab_locks);
-       ptr; ptr = hash_nextentry(&htab_locks)) {
+  lock_list **locks, *lk;
+  bool first = 1;
+  int nlocks = 0, n;
+
+  locks = mush_calloc(htab_locks.entries, sizeof(lock_list), "lock.list");
+
+  for (lk = hash_firstentry(&htab_locks); lk;
+       lk = hash_nextentry(&htab_locks)) {
     /* Skip those that don't match */
-    if (name && !string_prefix(ptr->type, name))
+    if (name && !string_prefix(lk->type, name))
       continue;
+    locks[nlocks++] = lk;
+  }
+
+  qsort(locks, nlocks, sizeof lk, lock_compare);
+
+  for (n = 0; n < nlocks; n += 1) {
     if (first) {
       first = 0;
     } else {
-      safe_chr(' ', rbuff, &rp);
+      safe_chr(' ', buff, bp);
     }
-    safe_str(ptr->type, rbuff, &rp);
+    safe_str(strupper(locks[n]->type), buff, bp);
   }
-  *rp = '\0';
-  /* We strupper it for consistency with the other
-   * @list/foo and list(foo)s. */
-  safe_str(strupper(rbuff), buff, bp);
+
+  mush_free(locks, "lock.list");
 }
 
 /** User interface to list locks.
@@ -199,7 +180,6 @@ do_list_locks(dbref player, const char *arg, int lc, const char *label)
   notify_format(player, "%s: %s", label, lc ? strlower(buff) : buff);
 }
 
-
 /** Return a list of lock flag characters.
  * \param ll pointer to a lock.
  * \return string of lock flag characters.
@@ -253,7 +233,6 @@ lock_flags_long(lock_list *ll)
   return privs_to_string(lock_privs, L_FLAGS(ll));
 }
 
-
 static int
 string_to_lockflag(dbref player, char const *p, privbits *flag)
 {
@@ -272,7 +251,7 @@ void
 init_locks(void)
 {
   lock_list *ll;
-  st_init(&lock_names);
+  st_init(&lock_names, "LockNameTree");
 
   hashinit(&htab_locks, 25);
 
@@ -282,7 +261,6 @@ init_locks(void)
   local_locks();
 }
 
-
 /** Add a new lock to the table.
  * \param name The name of the lock
  * \param flags The default flags.
@@ -293,7 +271,7 @@ define_lock(lock_type name, privbits flags)
   lock_list *newlock;
 
   newlock = mush_malloc(sizeof *newlock, "lock");
-  newlock->type = mush_strdup(strupper(name), "lock.name");
+  newlock->type = strupper_a(name, "lock.name");
   newlock->flags = flags;
   newlock->creator = GOD;
   newlock->key = TRUE_BOOLEXP;
@@ -312,14 +290,14 @@ can_write_lock(dbref player, dbref thing, lock_list *lock)
     return 1;
   if (L_FLAGS(lock) & LF_WIZARD)
     return 0;
-  if (L_FLAGS(lock) & LF_OWNER)
-    return player == Owner(thing);
-  if (L_FLAGS(lock) & LF_LOCKED)
-    return Owner(player) == lock->creator;
+  if ((L_FLAGS(lock) & LF_OWNER) && player != Owner(thing))
+    return 0;
+  if ((L_FLAGS(lock) & LF_LOCKED) && player != lock->creator &&
+      Owner(player) != lock->creator)
+    return 0;
   return 1;
 }
 
-
 static lock_list *
 next_free_lock(const void *hint)
 {
@@ -351,7 +329,7 @@ getlock(dbref thing, lock_type type)
     return L_KEY(ll);
 }
 
-/** Given a lock type, find a lock without checking parents. 
+/** Given a lock type, find a lock without checking parents.
  * \param thing object on which lock is to be found.
  * \param type type of lock to find.
  * \return pointer to boolexp of lock.
@@ -413,7 +391,6 @@ getlockstruct_noparent(dbref thing, lock_type type)
   return NULL;
 }
 
-
 /** Determine if a lock type is one of the standard types or not.
  * \param type type of lock to check.
  * \return canonical lock type or NULL
@@ -436,7 +413,7 @@ match_lock(lock_type type)
 const lock_list *
 get_lockproto(lock_type type)
 {
-  return hashfind(type, &htab_locks);
+  return hashfind(strupper(type), &htab_locks);
 }
 
 /** Add a lock to an object (primitive).
@@ -457,7 +434,6 @@ int
 add_lock(dbref player, dbref thing, lock_type type, boolexp key, privbits flags)
 {
   lock_list *ll, **t;
-  lock_type real_type = type;
 
   if (!GoodObject(thing)) {
     return 0;
@@ -482,7 +458,7 @@ add_lock(dbref player, dbref thing, lock_type type, boolexp key, privbits flags)
       /* Oh, this sucks */
       do_log(LT_ERR, 0, 0, "Unable to malloc memory for lock_list!");
     } else {
-      real_type = st_insert(type, &lock_names);
+      lock_type real_type = st_insert(type, &lock_names);
       ll->type = real_type;
       ll->key = key;
       ll->creator = player;
@@ -618,25 +594,24 @@ free_locks(lock_list *ll)
   }
 }
 
-
-/** Check to see that the lock type is a valid type.
- * If it's not in our lock table, it's not valid,
- * unless it begins with 'user:' or an abbreviation thereof,
- * in which case the lock type is the part after the :.
- * As an extra check, we don't allow '|' in lock names because it
- * will confuse our db-reading routines.
- *
- *
+/** Check to see that the lock type is a valid type. Valid types are either:
+ * 1) in the lock table of standard lock types,
+ * 2) prefixed with "user:", a valid attr name and don't contain '|' chars
+ * 3) the name of an already-existing lock set on 'thing'
+ * It was previously stated that '|' in lock names interfered with the db
+ * reading routines; I (MG) don't believe that's the case any longer, though.
  * \param player the enactor, for notification.
  * \param thing object on which to check the lock.
+ * \param silent skip specific error notifications to player?
  * \param name name of lock type.
- * \return lock type, or NULL.
+ * \retval NULL invalid lock type
+ * \retval locktype the name of the valid lock type, minus user: prefix
  */
-static lock_type
-check_lock_type(dbref player, dbref thing, lock_type name)
+lock_type
+check_lock_type(dbref player, dbref thing, lock_type name, bool silent)
 {
   lock_type ll;
-  char user_name[BUFFER_LEN];
+  char *user_name;
   char *colon;
 
   /* Special-case for basic locks. */
@@ -651,27 +626,28 @@ check_lock_type(dbref player, dbref thing, lock_type name)
   /* If the lock is set, it's allowed, whether it exists normally or not. */
   if (getlock(thing, name) != TRUE_BOOLEXP)
     return name;
-  /* Check to see if it's a well-formed user-defined lock. */
 
+  /* Check to see if it's a well-formed user-defined lock. */
   if (!string_prefix(name, "User:")) {
-    notify(player, T("Unknown lock type."));
+    if (!silent)
+      notify(player, T("Unknown lock type."));
     return NULL;
   }
   if (strchr(name, '|')) {
-    notify(player, T("The character \'|\' may not be used in lock names."));
+    if (!silent)
+      notify(player, T("The character \'|\' may not be used in lock names."));
     return NULL;
   }
-  mush_strncpy(user_name, name, BUFFER_LEN);
-  colon = strchr(user_name, ':');
-  if (colon)                    /* Should always be true */
-    *colon = '\0';
+  colon = strchr(name, ':') + 1;
+  user_name = strupper(colon);
 
   if (!good_atr_name(user_name)) {
-    notify(player, T("That is not a valid lock name."));
+    if (!silent)
+      notify(player, T("That is not a valid lock name."));
     return NULL;
   }
 
-  return strchr(name, ':') + 1;
+  return colon;
 }
 
 /** Unlock a lock (user interface).
@@ -694,15 +670,17 @@ do_unlock(dbref player, const char *name, lock_type type)
     return;
   }
   if ((thing = match_controlled(player, name)) != NOTHING) {
-    if ((real_type = check_lock_type(player, thing, type)) != NULL) {
+    if ((real_type = check_lock_type(player, thing, type, 0)) != NULL) {
       if (getlock(thing, real_type) == TRUE_BOOLEXP) {
         if (!AreQuiet(player, thing))
           notify_format(player, T("%s(%s) - %s (already) unlocked."),
-                        Name(thing), unparse_dbref(thing), real_type);
+                        AName(thing, AN_SYS, NULL), unparse_dbref(thing),
+                        real_type);
       } else if (delete_lock(player, thing, real_type)) {
         if (!AreQuiet(player, thing))
-          notify_format(player, T("%s(%s) - %s unlocked."), Name(thing),
-                        unparse_dbref(thing), real_type);
+          notify_format(player, T("%s(%s) - %s unlocked."),
+                        AName(thing, AN_SYS, NULL), unparse_dbref(thing),
+                        real_type);
         if (!IsPlayer(thing))
           ModTime(thing) = mudtime;
       } else
@@ -761,17 +739,18 @@ do_lock(dbref player, const char *name, const char *keyname, lock_type type)
   if (key == TRUE_BOOLEXP) {
     notify(player, T("I don't understand that key."));
   } else {
-    if ((real_type = check_lock_type(player, thing, type)) != NULL) {
+    if ((real_type = check_lock_type(player, thing, type, 0)) != NULL) {
       /* everything ok, do it */
       if (add_lock(player, thing, real_type, key, LF_DEFAULT)) {
         if (!AreQuiet(player, thing))
-          notify_format(player, T("%s(%s) - %s locked."), Name(thing),
-                        unparse_dbref(thing), real_type);
+          notify_format(player, T("%s(%s) - %s locked."),
+                        AName(thing, AN_SYS, NULL), unparse_dbref(thing),
+                        real_type);
         if (!IsPlayer(thing))
           ModTime(thing) = mudtime;
       } else {
         notify(player, T("Permission denied."));
-        free_boolexp(key);
+        /*  Done by a failed add_lock()  // free_boolexp(key); */
       }
     } else
       free_boolexp(key);
@@ -793,65 +772,51 @@ clone_locks(dbref player, dbref orig, dbref clone)
   }
 }
 
-
 /** Evaluate a lock.
  * Evaluate lock ltype on thing for player.
  * \param player dbref attempting to pass the lock.
  * \param thing object containing the lock.
  * \param ltype type of lock to check.
+ * \param pe_info the pe_info to use when evaluating softcode in the lock
  * \retval 1 player passes the lock.
  * \retval 0 player does not pass the lock.
  */
 int
-eval_lock(dbref player, dbref thing, lock_type ltype)
+eval_lock_with(dbref player, dbref thing, lock_type ltype, NEW_PE_INFO *pe_info)
 {
   boolexp b = getlock(thing, ltype);
   /* Prevent overwriting a static buffer in unparse_boolexp() */
   if (!unparsing_boolexp)
     log_activity(LA_LOCK, thing, unparse_boolexp(player, b, UB_DBREF));
-  return eval_boolexp(player, b, thing);
+  return eval_boolexp(player, b, thing, pe_info);
 }
 
-/** Evaluate a lock.
- * Evaluate lock ltype on thing for player, passing dbrefs for %0/%1.
- * \param player dbref attempting to pass the lock.
- * \param thing object containing the lock.
- * \param ltype type of lock to check.
- * \retval 1 player passes the lock.
- * \retval 0 player does not pass the lock.
+/* eval_lock(player,thing,ltype) is #defined to
+ * eval_lock_with(player,thing,ltype,NULL) */
+
+/** Evaluate a lock, saving/clearing the env (%0-%9) and qreg (%q*) first,
+ ** and restoring them after.
  */
 int
-eval_lock_with(dbref player, dbref thing, lock_type ltype, dbref env0,
-               dbref env1)
+eval_lock_clear(dbref player, dbref thing, lock_type ltype,
+                NEW_PE_INFO *pe_info)
 {
-  char *myenv[10] = { NULL };
-  char e0[SBUF_LEN], e1[SBUF_LEN], *ep;
-  char *preserves[10];
-  int result;
-  boolexp b;
+  if (!pe_info)
+    return eval_lock_with(player, thing, ltype, NULL);
+  else {
+    PE_REGS *pe_regs;
+    int result;
 
-  if (env0 != NOTHING) {
-    ep = e0;
-    safe_dbref(env0, e0, &ep);
-    *ep = '\0';
-    myenv[0] = e0;
+    pe_regs = pe_regs_localize(pe_info, PE_REGS_ISOLATE, "eval_lock_clear");
+
+    /* Run the lock */
+    result = eval_lock_with(player, thing, ltype, pe_info);
+
+    /* Restore q-regs */
+    pe_regs_restore(pe_info, pe_regs);
+    pe_regs_free(pe_regs);
+    return result;
   }
-
-  if (env1 != NOTHING) {
-    ep = e1;
-    safe_dbref(env1, e1, &ep);
-    *ep = '\0';
-    myenv[1] = e1;
-  }
-
-  save_global_env("eval_lock_save", preserves);
-  restore_global_env("eval_lock", myenv);
-
-  b = getlock(thing, ltype);
-  log_activity(LA_LOCK, thing, unparse_boolexp(player, b, UB_DBREF));
-  result = eval_boolexp(player, b, thing);
-  restore_global_env("eval_lock_save", preserves);
-  return result;
 }
 
 /** Active a lock's failure attributes.
@@ -871,12 +836,18 @@ fail_lock(dbref player, dbref thing, lock_type ltype, const char *def,
   char atr[BUFFER_LEN];
   char oatr[BUFFER_LEN];
   char aatr[BUFFER_LEN];
+  char realdef[BUFFER_LEN];
   char *bp;
 
+  if (def)
+    strcpy(realdef, def); /* Because a lot of default msgs use tprintf */
+  else
+    realdef[0] = '\0';
+
   /* Find the lock's failure attribute, if it's there */
   for (lm = lock_msgs; lm->type; lm++) {
     if (!strcmp(lm->type, ltype))
-       break;
+      break;
   }
   if (lm->type) {
     strcpy(atr, lm->failbase);
@@ -903,10 +874,9 @@ fail_lock(dbref player, dbref thing, lock_type ltype, const char *def,
   upcasestr(atr);
   upcasestr(oatr);
   upcasestr(aatr);
-  return did_it(player, thing, atr, def, oatr, NULL, aatr, loc);
+  return did_it(player, thing, atr, realdef, oatr, NULL, aatr, loc, AN_SYS);
 }
 
-
 /** Determine if a lock is visual.
  * \param thing object containing the lock.
  * \param ltype type of lock to check.
@@ -976,7 +946,7 @@ do_lset(dbref player, char *what, char *flags)
     L_FLAGS(l) |= flag;
 
   if (!Quiet(player) && !(Quiet(thing) && (Owner(thing) == player)))
-    notify_format(player, "%s/%s - %s.", Name(thing), L_TYPE(l),
+    notify_format(player, "%s/%s - %s.", AName(thing, AN_SYS, NULL), L_TYPE(l),
                   unset ? T("lock flags unset") : T("lock flags set"));
   if (!IsPlayer(thing))
     ModTime(thing) = mudtime;
@@ -995,23 +965,38 @@ check_zone_lock(dbref player, dbref zone, int noisy)
   if (key == TRUE_BOOLEXP) {
     add_lock(GOD, zone, Zone_Lock, parse_boolexp(zone, "=me", Zone_Lock),
              LF_DEFAULT);
-    if (noisy)
-      notify_format(player,
-                    T
-                    ("Unlocked zone %s - automatically zone-locking to itself"),
-                    unparse_object(player, zone));
+    if (noisy) {
+      notify_format(
+        player, T("Unlocked zone %s - automatically zone-locking to itself"),
+        unparse_object(player, zone, AN_UNPARSE));
+    }
+  } else if (!noisy) {
+    return;
   } else if (eval_lock(Location(player), zone, Zone_Lock)) {
     /* Does #0 and #2 pass it? If so, probably trivial elock */
     if (eval_lock(PLAYER_START, zone, Zone_Lock) &&
         eval_lock(MASTER_ROOM, zone, Zone_Lock)) {
-      if (noisy)
-        notify_format(player,
-                      T("Zone %s really should have a more secure zone-lock."),
-                      unparse_object(player, zone));
-    } else                      /* Probably inexact zone lock */
       notify_format(player,
-                    T
-                    ("Warning: Zone %s may have loose zone lock. Lock zones to =player, not player"),
-                    unparse_object(player, zone));
+                    T("Zone %s really should have a more secure zone-lock."),
+                    unparse_object(player, zone, AN_UNPARSE));
+    } else {
+      /* Probably inexact zone lock */
+      notify_format(player,
+                    T("Warning: Zone %s may have loose zone lock. Lock "
+                      "zones to =player, not player"),
+                    unparse_object(player, zone, AN_UNPARSE));
+    }
+  }
+}
+
+void
+purge_locks(void)
+{
+  dbref thing;
+
+  for (thing = 0; thing < db_top; thing++) {
+    lock_list *ll;
+    for (ll = Locks(thing); ll; ll = L_NEXT(ll))
+      L_KEY(ll) = cleanup_boolexp(L_KEY(ll));
   }
 }
