@@ -3,10 +3,7 @@
  *
  * \brief Malloc wrapper file.
  *
- * Three things in this file:
- *
- * -# It includes the body of csrimalloc.c if using a MALLOC_PACKAGE
- *     of 2 or 3.
+ * Two things in this file:
  *
  * -# It has the mush_FOO() wrapper functions for
  *     malloc()/calloc()/realloc()/free(). These are used to keep
@@ -27,9 +24,10 @@
  *     overhead.
  *
  */
-#include "config.h"
+
+#include "mymalloc.h"
+
 #include <stdlib.h>
-#include <stddef.h>
 #include <limits.h>
 #include <assert.h>
 #include <errno.h>
@@ -37,22 +35,29 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
 #ifdef WIN32
 #include <windows.h>
 #endif
-#include "options.h"
+
 #include "conf.h"
 #include "dbdefs.h"
 #include "log.h"
-#include "externs.h"
-#include "getpgsiz.h"
-#include "mymalloc.h"
-#include "confmagic.h"
+#include "memcheck.h"
+#include "strutil.h"
+
+#ifdef WIN32
+#define SZT "I64u"
+#else
+#define SZT "zu"
+#endif
 
 /** A malloc wrapper that tracks type of allocation.
  * This should be used in preference to malloc() when possible,
  * to enable memory leak tracing with MEM_CHECK.
- * \param size bytes to allocate.
+ * \param bytes bytes to allocate.
  * \param check string to label allocation with.
  * \return allocated block of memory or NULL.
  */
@@ -60,10 +65,36 @@ void *
 mush_malloc(size_t bytes, const char *check)
 {
   void *ptr;
+
+#ifdef HAVE_SSE42
+  if (strcmp(check, "string") == 0 ||
+      strcmp(check, "descriptor_raw_input") == 0)
+    bytes += 16;
+#endif
+
   ptr = malloc(bytes);
   if (!ptr)
-    do_rawlog(LT_TRACE, "mush_malloc failed to malloc %lu bytes for %s",
-              (unsigned long) bytes, check);
+    do_rawlog(LT_TRACE, "mush_malloc failed to malloc %" SZT " bytes for %s",
+              bytes, check);
+  add_check(check);
+  return ptr;
+}
+
+/** Like mush_malloc(), but ensures the returned memory is
+ * zero'ed out.
+ *
+ * \param bytes amount of memory to allocate.
+ * \param check string to label allocation with.
+ * \return allocated block of memory or NULL.
+ */
+void *
+mush_malloc_zero(size_t bytes, const char *check)
+{
+  void *ptr = calloc(bytes, 1);
+  if (!ptr)
+    do_rawlog(LT_TRACE,
+              "mush_malloc_zero failed to allocate %" SZT " bytes for %s",
+              bytes, check);
   add_check(check);
   return ptr;
 }
@@ -83,8 +114,8 @@ mush_calloc(size_t count, size_t size, const char *check)
 
   ptr = calloc(count, size);
   if (!ptr)
-    do_rawlog(LT_TRACE, "mush_calloc failed to allocate %lu bytes for %s",
-              (unsigned long) (size * count), check);
+    do_rawlog(LT_TRACE, "mush_calloc failed to allocate %" SZT " bytes for %s",
+              size * count, check);
   add_check(check);
   return ptr;
 }
@@ -97,7 +128,6 @@ mush_calloc(size_t count, size_t size, const char *check)
  * \param check string to label with
  * \param filename file name it was called from
  * \param line line it called from
- * \param new pointer or NULL
  */
 void *
 mush_realloc_where(void *restrict ptr, size_t newsize,
@@ -128,6 +158,13 @@ void
 mush_free_where(void *restrict ptr, const char *restrict check,
                 const char *restrict filename, int line)
 {
+#ifdef DEBUG
+  if (strstr(check, "pcre")) {
+    do_rawlog(LT_TRACE,
+              "pcre allocation at %s:%d should be freed with pcre2_code_free",
+              filename, line);
+  }
+#endif
   del_check(check, filename, line);
   free(ptr);
   return;
@@ -144,32 +181,32 @@ struct slab_page_list {
 
 /** A struct that represents one page's worth of objects */
 struct slab_page {
-  int nalloced; /**< Number of objects allocated from this page */
-  int nfree; /**< Number of objects on this page's free list */
-  void *last_obj; /**< Pointer to last object in page. */
+  int nalloced;           /**< Number of objects allocated from this page */
+  int nfree;              /**< Number of objects on this page's free list */
+  void *last_obj;         /**< Pointer to last object in page. */
   struct slab_page *next; /**< Pointer to next allocated page */
-  struct slab_page_list *freelist; /**< Pointer to list of unallocated objects */
+  struct slab_page_list
+    *freelist; /**< Pointer to list of unallocated objects */
 };
 
 /** Struct for a slab allocator */
 struct slab {
-  char name[64]; /**< Name of the slab */
-  int item_size; /**< Size of the objects this slab returns */
-  int items_per_page; /**< Number of objects that fit into a page */
-  ptrdiff_t data_offset; /**< offset from start of the page where objects
-                      are allocated from. */
-  bool fill_strategy; /**< How to find empty nodes? true for
-                         FIRST_FIT, false for BEST_FIT. */
-  bool keep_last_empty; /**< False if empty pages are always deleted,
-                           true to keep an empty page if it is the
-                           only allocated page. */
-  int hintless_threshold; /**< See documentation for
-                             SLAB_HINTLESS_THRESHOLD option */
+  char name[64];           /**< Name of the slab */
+  int item_size;           /**< Size of the objects this slab returns */
+  int items_per_page;      /**< Number of objects that fit into a page */
+  ptrdiff_t data_offset;   /**< offset from start of the page where objects
+                        are allocated from. */
+  bool fill_strategy;      /**< How to find empty nodes? true for
+                              FIRST_FIT, false for BEST_FIT. */
+  bool keep_last_empty;    /**< False if empty pages are always deleted,
+                              true to keep an empty page if it is the
+                              only allocated page. */
+  int hintless_threshold;  /**< See documentation for
+                              SLAB_HINTLESS_THRESHOLD option */
   struct slab_page *slabs; /**< Pointer to the head of the list of
                               allocated pages. */
 };
 
-
 /** Create a new slab allocator.
  * \param name The name of the allocator
  * \param item_size The size of objects to be allocated. It must be smaller
@@ -183,7 +220,7 @@ slab_create(const char *name, size_t item_size)
   size_t pgsize, offset;
 
   sl = malloc(sizeof(struct slab));
-  pgsize = getpagesize();
+  pgsize = mush_getpagesize();
   offset = sizeof(struct slab_page);
   /* Start the objects 16-byte aligned */
   offset += sizeof(struct slab_page) % 16;
@@ -193,16 +230,18 @@ slab_create(const char *name, size_t item_size)
   sl->keep_last_empty = 0;
   sl->hintless_threshold = 0;
   sl->slabs = NULL;
-  if (item_size < SIZEOF_VOID_P)
-    item_size = SIZEOF_VOID_P;
+  if (item_size < sizeof(void *))
+    item_size = sizeof(void *);
   /* Align objects after the first with the size of a pointer */
-  item_size += item_size % SIZEOF_VOID_P;
+  item_size += item_size % sizeof(void *);
   sl->item_size = item_size;
   sl->items_per_page = (pgsize - offset) / item_size;
 
   if (item_size >= (pgsize - offset)) {
     do_rawlog(LT_TRACE,
-              "slab(%s): item_size of %lu bytes is too large for a pagesize of %lu bytes. Using malloc() for allocations from this slab.",
+              "slab(%s): item_size of %lu bytes is too large for a "
+              "pagesize of %lu bytes. Using malloc() for allocations "
+              "from this slab.",
               name, (unsigned long) item_size, (unsigned long) pgsize);
     sl->items_per_page = 0;
     return sl;
@@ -217,8 +256,8 @@ slab_create(const char *name, size_t item_size)
  * \param val Value to set option too (Not meaningful for all options)
  */
 void
-slab_set_opt(slab *sl, enum slab_options opt, int val
-             __attribute__ ((__unused__)))
+slab_set_opt(slab *sl, enum slab_options opt,
+             int val __attribute__((__unused__)))
 {
   if (!sl)
     return;
@@ -243,7 +282,7 @@ slab_set_opt(slab *sl, enum slab_options opt, int val
 
 /** Allocate a new page.
  * \param sl Allocator to create the page for.
- * \return new page, NOT linked into the allocator's list of pages 
+ * \return new page, NOT linked into the allocator's list of pages
  */
 static struct slab_page *
 slab_alloc_page(struct slab *sl)
@@ -253,7 +292,7 @@ slab_alloc_page(struct slab *sl)
   int n;
   int pgsize;
 
-  pgsize = getpagesize();
+  pgsize = mush_getpagesize();
 
 #ifdef HAVE_POSIX_MEMALIGN
   /* Used to use valloc() here, but on some systems, memory returned by
@@ -268,6 +307,7 @@ slab_alloc_page(struct slab *sl)
 #else
   page = malloc(pgsize);
 #endif
+  memset(page, 0, pgsize);
 
   sp = (struct slab_page *) page;
   sp->nfree = sl->items_per_page;
@@ -284,7 +324,8 @@ slab_alloc_page(struct slab *sl)
   sp->next = NULL;
 #ifdef SLAB_DEBUG
   do_rawlog(LT_TRACE,
-            "Allocating page starting at %p for slab(%s).\n\tFirst object allocated at %p, last object at %p",
+            "Allocating page starting at %p for slab(%s).\n\tFirst "
+            "object allocated at %p, last object at %p",
             (void *) sp, sl->name, (void *) (sp + sl->data_offset),
             sp->last_obj);
 #endif
@@ -292,7 +333,7 @@ slab_alloc_page(struct slab *sl)
   return sp;
 }
 
-/** Allocate a new object from a page 
+/** Allocate a new object from a page
  * \param where the page to allocate from.
  * \return pointer to object, or NULL if no room left on page
  */
@@ -361,10 +402,8 @@ slab_malloc(slab *sl, const void *hint)
     return slab_alloc_obj(last->next);
   } else {
     struct slab_page *page, *last = NULL;
-    ptrdiff_t pgsize;
     /* Okay. We have a hint for where to allocate the object. Find the
        page the hint is on. */
-    pgsize = getpagesize();
     for (page = sl->slabs; page; page = page->next) {
       if (hint > (void *) page && hint <= page->last_obj) {
         /* If there's space, use this page, otherwise, if using
@@ -384,7 +423,7 @@ slab_malloc(slab *sl, const void *hint)
       }
       last = page;
     }
-    /* This should never be reached, but handle it anyways. */
+/* This should never be reached, but handle it anyways. */
 #ifdef SLAB_DEBUG
     do_rawlog(LT_TRACE, "page hint %p not found in slab(%s)", (void *) hint,
               sl->name);
@@ -403,7 +442,6 @@ void
 slab_free(slab *sl, void *obj)
 {
   struct slab_page *page, *last;
-  ptrdiff_t pgsize;
 
   /* If objects are too big to fit in a single page, use plain free */
   if (sl->items_per_page == 0) {
@@ -412,7 +450,6 @@ slab_free(slab *sl, void *obj)
   }
 
   /* Find the page the object is on and push it into that page's free list */
-  pgsize = getpagesize();
   last = NULL;
   for (page = sl->slabs; page; page = page->next) {
     if (obj > (void *) page && obj <= page->last_obj) {
@@ -421,9 +458,10 @@ slab_free(slab *sl, void *obj)
       struct slab_page_list *scan;
       for (scan = page->freelist; scan; scan = scan->next)
         if (item == scan)
-          do_rawlog(LT_TRACE,
-                    "Attempt to free already free object %p from page %p of slab(%s)",
-                    (void *) item, (void *) page, sl->name);
+          do_rawlog(
+            LT_TRACE,
+            "Attempt to free already free object %p from page %p of slab(%s)",
+            (void *) item, (void *) page, sl->name);
 #endif
       item->next = page->freelist;
       page->freelist = item;
@@ -456,8 +494,7 @@ slab_free(slab *sl, void *obj)
     last = page;
   }
   /* Ooops. An object not allocated by this allocator! */
-  do_rawlog(LT_TRACE,
-            "Attempt to free object %p not allocated by slab(%s)",
+  do_rawlog(LT_TRACE, "Attempt to free object %p not allocated by slab(%s)",
             obj, sl->name);
 }
 
@@ -477,126 +514,61 @@ slab_destroy(slab *sl)
   free(sl);
 }
 
-/** Describe a slab for @list allocations
- * \param player who to display to
+/** Retrieve interesting stats about a slab.
  * \param sl the slab
  */
 void
-slab_describe(dbref player, slab *sl)
+slab_describe(const slab *sl, struct slab_stats *stats)
 {
-  struct slab_page *page;
-  int n = 0;
-  size_t allocated = 0, freed = 0;
-  int min_fill = INT_MAX, max_fill = 0;
-  int full = 0, under100 = 0, under75 = 0, under50 = 0, under25 = 0;
+  const struct slab_page *page;
 
   if (!sl)
     return;
 
+  memset(stats, 0, sizeof(*stats));
+  stats->name = sl->name;
+  stats->item_size = sl->item_size;
+  stats->items_per_page = sl->items_per_page;
+  stats->fill_strategy = sl->fill_strategy;
+  stats->min_fill = INT_MAX;
+  stats->max_fill = 0;
+
   for (page = sl->slabs; page; page = page->next) {
     double p;
-    n++;
-    allocated += page->nalloced;
-    freed += page->nfree;
-    if (page->nalloced > max_fill)
-      max_fill = page->nalloced;
-    if (page->nalloced < min_fill)
-      min_fill = page->nalloced;
+    stats->page_count++;
+    stats->allocated += page->nalloced;
+    stats->freed += page->nfree;
+    if (page->nalloced > stats->max_fill)
+      stats->max_fill = page->nalloced;
+    if (page->nalloced < stats->min_fill)
+      stats->min_fill = page->nalloced;
     p = (double) page->nalloced / (double) sl->items_per_page;
     if (p == 1.0)
-      full += 1;
+      stats->full += 1;
     else if (p > 0.75)
-      under100 += 1;
+      stats->under100 += 1;
     else if (p > 0.50)
-      under75 += 1;
+      stats->under75 += 1;
     else if (p > 0.25)
-      under50 += 1;
+      stats->under50 += 1;
     else
-      under25 += 1;
-  }
-
-  notify_format(player, "Allocator for %s:", sl->name);
-  notify_format(player,
-                "   object size (bytes): %-6d       objects per page: %-6d",
-                sl->item_size, sl->items_per_page);
-  notify_format(player,
-                "       allocated pages: %-6d      objects added via: %s", n,
-                sl->fill_strategy ? "first fit" : "best fit");
-  notify_format(player,
-                "     allocated objects: %-6ld           free objects: %-6ld",
-                (unsigned long) allocated, (unsigned long) freed);
-  if (allocated > 0) {
-    notify_format(player,
-                  " fewest allocs in page: %-6d    most allocs in page: %-6d",
-                  min_fill, max_fill);
-    notify_format(player,
-                  "    allocation average:%6.2f%%        pages 100%% full: %-6d",
-                  (((double) allocated / ((double) allocated + (double) freed))
-                   * 100.0), full);
-    notify_format(player,
-                  "       pages >75%% full: %-6d        pages >50%% full: %-6d",
-                  under100, under75);
-    notify_format(player,
-                  "       pages >25%% full: %-6d        pages <25%% full: %d",
-                  under50, under25);
+      stats->under25 += 1;
   }
 }
 
-
-extern slab *attrib_slab, *lock_slab, *boolexp_slab, *bvm_asmnode_slab,
-  *bvm_strnode_slab, *flag_slab, *player_dbref_slab,
-  *command_slab, *channel_slab, *chanuser_slab, *chanlist_slab, *mail_slab,
-  *text_block_slab, *function_slab, *memcheck_slab, *intmap_slab;
-
-#if COMPRESSION_TYPE == 1 || COMPRESSION_TYPE == 2
-extern slab *huffman_slab;
-#endif
-
-/** List information about slab allocators and memcheck data. Admin only. */
-void
-do_list_allocations(dbref player)
-{
-  if (!Hasprivs(player)) {
-    notify(player, T("Sorry."));
-    return;
-  }
-  slab_describe(player, attrib_slab);
-#ifdef DEBUG
-  /* These should always be 0. No need to display them most of the
-     time. */
-  slab_describe(player, boolexp_slab);
-  slab_describe(player, bvm_asmnode_slab);
-  slab_describe(player, bvm_strnode_slab);
-#endif
-  slab_describe(player, channel_slab);
-  slab_describe(player, chanlist_slab);
-  slab_describe(player, chanuser_slab);
-  slab_describe(player, command_slab);
-  slab_describe(player, flag_slab);
-  slab_describe(player, function_slab);
-#if COMPRESSION_TYPE == 1 || COMPRESSION_TYPE == 2
-  slab_describe(player, huffman_slab);
-#endif
-  slab_describe(player, lock_slab);
-  slab_describe(player, mail_slab);
-  slab_describe(player, memcheck_slab);
-  slab_describe(player, text_block_slab);
-  slab_describe(player, player_dbref_slab);
-  slab_describe(player, intmap_slab);
-
-  if (options.mem_check) {
-    notify(player, "malloc allocations:");
-    list_mem_check(player);
-  }
-}
-
-#ifdef WIN32
-/** Windows version of getpagesize() */
-unsigned int
-getpagesize_win32(void)
+/** Return the memory page size */
+int
+mush_getpagesize(void)
 {
+#if defined(WIN32)
   SYSTEM_INFO si;
   GetSystemInfo(&si);
   return si.dwPageSize;
-}
+#elif defined(HAVE_GETPAGESIZE)
+  return getpagesize();
+#elif defined(HAVE_SYSCONF)
+  return sysconf(_SC_PAGESIZE);
+#else
+  return 4096;
 #endif
+}
