@@ -11,28 +11,20 @@
  */
 
 #include "copyrite.h"
-#include "config.h"
 
-#ifdef HAS_OPENSSL
 #include <stdio.h>
 #include <stdarg.h>
-#ifdef I_SYS_TYPES
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
 #ifdef WIN32
-#define FD_SETSIZE 256
+#include <winsock2.h>
 #include <windows.h>
-#include <winsock.h>
 #include <io.h>
-#define EWOULDBLOCK WSAEWOULDBLOCK
-#define MAXHOSTNAMELEN 32
-#define LC_MESSAGES 6
+#include <ntstatus.h>
 void shutdown_checkpoint(void);
-#else                           /* !WIN32 */
-#ifdef I_SYS_FILE
-#include <sys/file.h>
-#endif
-#ifdef I_SYS_TIME
+#else /* !WIN32 */
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #ifdef TIME_WITH_SYS_TIME
 #include <time.h>
@@ -42,56 +34,50 @@ void shutdown_checkpoint(void);
 #endif
 #include <sys/ioctl.h>
 #include <errno.h>
-#ifdef I_SYS_SOCKET
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
-#ifdef I_NETINET_IN
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
-#ifdef I_NETDB
-#include <netdb.h>
-#endif
-#ifdef I_SYS_PARAM
-#include <sys/param.h>
-#endif
-#ifdef I_SYS_STAT
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
-#endif                          /* !WIN32 */
+#endif /* !WIN32 */
 #include <fcntl.h>
 #include <ctype.h>
 #include <signal.h>
 #include <string.h>
 #include <stdlib.h>
-#ifdef I_SYS_SELECT
+#ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
-#ifdef I_UNISTD
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include <stdio.h>
 
-#include <openssl/ssl.h>
+#include <openssl/bn.h>
 #include <openssl/err.h>
 #include <openssl/dh.h>
+#include <openssl/ec.h>
 #include <openssl/evp.h>
 #include <openssl/rand.h>
 
-#include "conf.h"
-#include "mysocket.h"
-#include "externs.h"
 #include "myssl.h"
-#include "log.h"
+#include "pcg_basic.h"
+#include "conf.h"
 #include "parse.h"
-#include "SFMT.h"
+#include "wait.h"
 #include "confmagic.h"
 
-#define MYSSL_RB        0x1     /**< Read blocked (on read) */
-#define MYSSL_WB        0x2     /**< Write blocked (on write) */
-#define MYSSL_RBOW      0x4     /**< Read blocked (on write) */
-#define MYSSL_WBOR      0x8     /**< Write blocked (on read) */
-#define MYSSL_ACCEPT    0x10    /**< We need to call SSL_accept (again) */
-#define MYSSL_VERIFIED  0x20    /**< This is an authenticated connection */
-#define MYSSL_HANDSHAKE 0x40    /**< We need to call SSL_do_handshake */
+#define MYSSL_RB 0x1         /**< Read blocked (on read) */
+#define MYSSL_WB 0x2         /**< Write blocked (on write) */
+#define MYSSL_RBOW 0x4       /**< Read blocked (on write) */
+#define MYSSL_WBOR 0x8       /**< Write blocked (on read) */
+#define MYSSL_ACCEPT 0x10    /**< We need to call SSL_accept (again) */
+#define MYSSL_VERIFIED 0x20  /**< This is an authenticated connection */
+#define MYSSL_HANDSHAKE 0x40 /**< We need to call SSL_do_handshake */
 
 #undef MYSSL_DEBUG
 #ifdef MYSSL_DEBUG
@@ -101,23 +87,125 @@ void shutdown_checkpoint(void);
 #endif
 
 static void ssl_errordump(const char *msg);
-static int client_verify_callback(int preverify_ok, X509_STORE_CTX * x509_ctx);
-static DH *get_dh1024(void);
-
+static int client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx);
+static DH *get_dh2048(void);
 
 static BIO *bio_err = NULL;
 static SSL_CTX *ctx = NULL;
 
+static const char *
+time_string(void)
+{
+  static char buffer[100];
+  time_t now;
+  struct tm *ltm;
+
+  now = time(NULL);
+  ltm = localtime(&now);
+  strftime(buffer, 100, "[%Y-%m-%d %H:%M:%S]", ltm);
+
+  return buffer;
+}
+
+/** Generate 128 bits of random noise for seeding RNGs. Attempts to
+ * use various OS-specific sources of random bits, with a fallback
+ * based on time and pid. */
+void
+generate_seed(uint64_t seeds[])
+{
+  bool seed_generated = false;
+  static int stream_count = 0;
+  int len = sizeof(uint64_t) * 2;
+
+#ifdef HAVE_GETENTROPY
+  /* On OpenBSD and up to date Linux, use getentropy() to avoid the
+     open/read/close sequence with /dev/urandom */
+  if (!seed_generated && getentropy(seeds, len) == 0) {
+    fprintf(stderr, "%s Seeded RNG with getentropy()\n", time_string());
+    seed_generated = true;
+  }
+#endif
+
+#ifdef HAVE_ARC4RANDOM_BUF
+  /* Most (all?) of the BSDs have this seeder. Use it for the reasons
+     above. Also available on Linux with libbsd, but we don't check
+     for that. */
+  if (!seed_generated) {
+    arc4random_buf(seeds, len);
+    fprintf(stderr, "%s Seeded RNG with arc4random\n", time_string());
+    seed_generated = true;
+  }
+#endif
+
+#ifdef WIN32
+  if (!seed_generated) {
+    /* Use the Win32 bcrypto RNG interface */
+    if (BCryptGenRandom(NULL, (PUCHAR) seeds, len,
+                        BCRYPT_USE_SYSTEM_PREFERRED_RNG) == STATUS_SUCCESS) {
+      fprintf(stderr, "%s Seeding RNG with BCryptGenRandom()\n", time_string());
+      seed_generated = true;
+    }
+  }
+#endif
+
+#ifdef HAVE_DEV_URANDOM
+  if (!seed_generated) {
+    /* Seed from /dev/urandom if available */
+    int fd;
+
+    fd = open("/dev/urandom", O_RDONLY);
+    if (fd >= 0) {
+      int r = read(fd, (void *) seeds, len);
+      close(fd);
+      if (r != len) {
+        fprintf(stderr,
+                "%s Couldn't read from /dev/urandom! Resorting to normal "
+                "seeding method.\n",
+                time_string());
+      } else {
+        fprintf(stderr, "%s Seeding RNG with /dev/urandom\n", time_string());
+        seed_generated = true;
+      }
+    } else {
+      fprintf(stderr,
+              "%s Couldn't open /dev/urandom to seed random number "
+              "generator. Resorting to normal seeding method.\n",
+              time_string());
+    }
+  }
+#endif
+
+  if (!seed_generated) {
+    /* Default seeder. Pick a seed that's slightly random */
+#ifdef WIN32
+    seeds[0] = (uint64_t) time(NULL);
+    seeds[1] = (uint64_t) GetCurrentProcessId() + stream_count;
+#else
+    seeds[0] = (uint64_t) time(NULL);
+    seeds[1] = (uint64_t) getpid() + stream_count;
+#endif
+    stream_count += 1;
+  }
+}
+
 /** Initialize the SSL context.
  * \return pointer to SSL context object.
  */
 SSL_CTX *
-ssl_init(char *private_key_file, char *ca_file, int req_client_cert)
+ssl_init(char *private_key_file, char *certificate_file, char *ca_file, char *ca_dir,
+         int req_client_cert)
 {
-  SSL_METHOD *meth;
-  unsigned char context[128];
-  DH *dh;
+  const SSL_METHOD
+    *meth; /* If this const gives you a warning, you're
+              using an old version of OpenSSL. Walker, this means you! */
+  /* uint8_t context[128]; */
   unsigned int reps = 1;
+  pcg32_random_t rand_state;
+  uint64_t seeds[2];
+  bool seeded = false;
+  
+  bool foundKey = false;
+  bool foundCert = false;
 
   if (!bio_err) {
     if (!SSL_library_init())
@@ -127,54 +215,104 @@ ssl_init(char *private_key_file, char *ca_file, int req_client_cert)
     bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);
   }
 
-  do_rawlog(LT_ERR, "Seeding OpenSSL random number pool.");
+  lock_file(stderr);
+  fprintf(stderr, "%s Seeding OpenSSL random number pool.\n", time_string());
+  unlock_file(stderr);
   while (!RAND_status()) {
     /* At this point, a system with /dev/urandom or a EGD file in the usual
-       places will have enough entropy. Otherwise, be lazy and use random numbers
-       until it's satisfied. */
-    uint32_t gibberish[4];
+       places will have enough entropy. Otherwise, be lazy and use random
+       numbers until it's satisfied. */
+    uint32_t gibberish[8];
     int n;
 
-    for (n = 0; n < 4; n++)
-      gibberish[n] = gen_rand32();
+    if (!seeded) {
+      generate_seed(seeds);
+      pcg32_srandom_r(&rand_state, seeds[0], seeds[1]);
+      seeded = 1;
+    }
+
+    for (n = 0; n < 8; n++)
+      gibberish[n] = pcg32_random_r(&rand_state);
 
     RAND_seed(gibberish, sizeof gibberish);
 
     reps += 1;
   }
 
-  do_rawlog(LT_ERR, "Seeded after %u %s.", reps, reps > 1 ? "cycles" : "cycle");
-
+  lock_file(stderr);
+  fprintf(stderr, "%s Seeded after %u %s.\n", time_string(), reps,
+          reps > 1 ? "cycles" : "cycle");
+  unlock_file(stderr);
 
   /* Set up SIGPIPE handler here? */
 
   /* Create context */
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+  meth = TLS_server_method();
+#else
   meth = SSLv23_server_method();
+#endif
   ctx = SSL_CTX_new(meth);
+  SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
 
   /* Load keys/certs */
   if (private_key_file && *private_key_file) {
-    if (!SSL_CTX_use_certificate_chain_file(ctx, private_key_file)) {
-      ssl_errordump
-        ("Unable to load server certificate - only anonymous ciphers supported.");
+    if (SSL_CTX_use_certificate_chain_file(ctx, private_key_file)) {
+      foundCert = true;
     }
-    if (!SSL_CTX_use_PrivateKey_file(ctx, private_key_file, SSL_FILETYPE_PEM)) {
-      ssl_errordump
-        ("Unable to load private key - only anonymous ciphers supported.");
+    if (SSL_CTX_use_PrivateKey_file(ctx, private_key_file, SSL_FILETYPE_PEM)) {
+      foundKey = true;
     }
   }
 
+  /* Load keys/certs from alternative file */
+  /* Should really load the key from one file and the cert from another */
+  /* but since the older behavior is to load them from the same file we */
+  /* will try to load both from both files to avoid user confusion */
+  if (certificate_file && *certificate_file) {
+    if (!foundCert) {
+      if (SSL_CTX_use_certificate_chain_file(ctx, certificate_file)) {
+        foundCert = true;
+      }
+    }
+    
+    if (!foundKey) {
+      if (SSL_CTX_use_PrivateKey_file(ctx, certificate_file, SSL_FILETYPE_PEM)) {
+        foundKey = true;
+      }
+    }
+  }
+
+  /* Add a log entry if we didn't find a certificate or key in either file */
+  if (!foundCert) {
+    ssl_errordump("Unable to load server certificate - only anonymous "
+                  "ciphers supported.");
+  }
+  
+  if (!foundKey) {
+    ssl_errordump("Unable to load private key - only anonymous ciphers supported.");
+  }
+  
   /* Load trusted CAs */
-  if (ca_file && *ca_file) {
-    if (!SSL_CTX_load_verify_locations(ctx, ca_file, NULL)) {
+  if ((ca_file && *ca_file) || (ca_dir && *ca_dir)) {
+    if (!SSL_CTX_load_verify_locations(ctx,
+                                       (ca_file && *ca_file) ? ca_file : NULL,
+                                       (ca_dir && *ca_dir) ? ca_dir : NULL)) {
       ssl_errordump("Unable to load CA certificates");
-    } else {
+    }
+    {
+      STACK_OF(X509_NAME) *certs = NULL;
+      if (ca_file && *ca_file)
+        certs = SSL_load_client_CA_file(ca_file);
+      if (certs)
+        SSL_CTX_set_client_CA_list(ctx, certs);
+
       if (req_client_cert)
         SSL_CTX_set_verify(ctx,
                            SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                            client_verify_callback);
       else
-        SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, client_verify_callback);
+        SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, client_verify_callback);
 #if (OPENSSL_VERSION_NUMBER < 0x0090600fL)
       SSL_CTX_set_verify_depth(ctx, 1);
 #endif
@@ -182,33 +320,43 @@ ssl_init(char *private_key_file, char *ca_file, int req_client_cert)
   }
 
   SSL_CTX_set_options(ctx, SSL_OP_SINGLE_DH_USE | SSL_OP_ALL);
-  SSL_CTX_set_mode(ctx,
-                   SSL_MODE_ENABLE_PARTIAL_WRITE |
-                   SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
+  SSL_CTX_set_mode(ctx, SSL_MODE_ENABLE_PARTIAL_WRITE |
+                          SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
 
-  /* Set up DH callback */
-  dh = get_dh1024();
-  SSL_CTX_set_tmp_dh(ctx, dh);
-  /* The above function makes a private copy of this */
-  DH_free(dh);
+  /* Set up DH key */
+  {
+    DH *dh;
+    dh = get_dh2048();
+    SSL_CTX_set_tmp_dh(ctx, dh);
+    DH_free(dh);
+  }
+
+#ifdef NID_X9_62_prime256v1
+  /* Set up ECDH key */
+  {
+    EC_KEY *ecdh = NULL;
+    ecdh = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
+    SSL_CTX_set_tmp_ecdh(ctx, ecdh);
+    EC_KEY_free(ecdh);
+  }
+#endif
 
   /* Set the cipher list to the usual default list, except that
    * we'll allow anonymous diffie-hellman, too.
    */
-  SSL_CTX_set_cipher_list(ctx, "ALL:ADH:RC4+RSA:+SSLv2:@STRENGTH");
+  SSL_CTX_set_cipher_list(ctx, "ALL:ECDH:ADH:!LOW:!MEDIUM:@STRENGTH");
 
   /* Set up session cache if we can */
-  strncpy((char *) context, MUDNAME, 128);
-  SSL_CTX_set_session_id_context(ctx, context, u_strlen(context));
-
-  /* Load hash algorithms */
-  OpenSSL_add_all_digests();
+  /*
+     strncpy((char *) context, MUDNAME, 128);
+     SSL_CTX_set_session_id_context(ctx, context, strlen(context));
+   */
 
   return ctx;
 }
 
 static int
-client_verify_callback(int preverify_ok, X509_STORE_CTX * x509_ctx)
+client_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 {
   char buf[256];
   X509 *err_cert;
@@ -220,68 +368,110 @@ client_verify_callback(int preverify_ok, X509_STORE_CTX * x509_ctx)
 
   X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);
   if (!preverify_ok) {
-    do_log(LT_ERR, 0, 0, "verify error:num=%d:%s:depth=%d:%s\n", err,
-           X509_verify_cert_error_string(err), depth, buf);
+    lock_file(stderr);
+    fprintf(stderr, "%s verify error:num=%d:%s:depth=%d:%s\n", time_string(),
+            err, X509_verify_cert_error_string(err), depth, buf);
     if (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT) {
-      X509_NAME_oneline(X509_get_issuer_name(x509_ctx->current_cert), buf, 256);
-      do_log(LT_ERR, 0, 0, "issuer= %s\n", buf);
+      X509_NAME_oneline(
+        X509_get_issuer_name(X509_STORE_CTX_get_current_cert(x509_ctx)), buf,
+        256);
+      fprintf(stderr, "%s issuer= %s\n", time_string(), buf);
     }
+    unlock_file(stderr);
     return preverify_ok;
   }
   /* They've passed the preverification */
-  /* if there are contents of the cert we wanted to verify, we'd do it here. 
+  /* if there are contents of the cert we wanted to verify, we'd do it here.
    */
   return preverify_ok;
 }
 
 static DH *
-get_dh1024(void)
+get_dh2048(void)
 {
-  static unsigned char dh1024_p[] = {
-    0xB6, 0xBC, 0x30, 0x5B, 0xB4, 0xE5, 0x96, 0x62, 0x3F, 0x85, 0x5B, 0x1F,
-    0x88, 0xD1, 0x12, 0xE1, 0x1D, 0x27, 0x69, 0x63, 0xAD, 0xB3, 0x4D, 0x23,
-    0xB8, 0x4B, 0x1A, 0x90, 0xA6, 0x89, 0xD8, 0x5D, 0xFA, 0xF5, 0x8F, 0xFF,
-    0xFF, 0xF4, 0x54, 0x3B, 0xCD, 0x5C, 0xAA, 0x79, 0x8B, 0x14, 0xBB, 0x84,
-    0xAC, 0xEE, 0x94, 0x47, 0x76, 0xEC, 0x46, 0x75, 0x26, 0x48, 0x8C, 0x06,
-    0x55, 0x27, 0x7F, 0xC0, 0xF1, 0xE8, 0x1F, 0xD2, 0xE4, 0x55, 0xAE, 0x78,
-    0x11, 0x6E, 0xF1, 0x3B, 0xCD, 0x55, 0xE8, 0x17, 0xE9, 0x15, 0x7B, 0x05,
-    0x91, 0x28, 0x9D, 0xD3, 0x40, 0x2E, 0x34, 0x03, 0x04, 0x2B, 0x2D, 0xC5,
-    0x5C, 0x67, 0xC5, 0xF4, 0x28, 0x8E, 0x16, 0xAC, 0xDD, 0x68, 0x43, 0x66,
-    0x51, 0xC1, 0x6F, 0x54, 0xB9, 0x22, 0xD8, 0x1A, 0x39, 0x6B, 0x0A, 0xC1,
-    0x20, 0x5A, 0x9D, 0x31, 0x30, 0xE4, 0x0B, 0xC3,
+  static const uint8_t dh2048_p[] = {
+    0x8C, 0x9A, 0x5A, 0x28, 0xBF, 0x13, 0x24, 0xC0, 0xD3, 0x7A, 0x73, 0xC3,
+    0x87, 0x5B, 0x80, 0x81, 0xE8, 0xF3, 0x7B, 0xF6, 0xF7, 0x18, 0x71, 0xF9,
+    0xBB, 0x5B, 0x88, 0x21, 0xAB, 0x63, 0xF6, 0x82, 0xA6, 0xEC, 0xD7, 0x04,
+    0x25, 0xDC, 0x64, 0x75, 0x00, 0x49, 0x2C, 0x13, 0x04, 0x4F, 0xCF, 0xF9,
+    0x06, 0xE0, 0x4D, 0x23, 0xB8, 0x7C, 0xD8, 0x29, 0x59, 0x6F, 0x69, 0xCC,
+    0x41, 0x1F, 0x45, 0xF8, 0x25, 0xC8, 0x72, 0xF4, 0xC8, 0x37, 0x3C, 0x30,
+    0xC2, 0x5A, 0xF3, 0x14, 0x43, 0x98, 0x4F, 0x99, 0x12, 0xBC, 0x68, 0x7E,
+    0x20, 0x24, 0xAA, 0x8B, 0xBA, 0x87, 0x32, 0xBC, 0x4B, 0xF3, 0x16, 0x25,
+    0xEE, 0xE5, 0xEB, 0x47, 0xED, 0xB2, 0x7D, 0x8F, 0x4F, 0xC8, 0xFB, 0x58,
+    0x3D, 0x2E, 0xF6, 0x54, 0xF4, 0xDA, 0xD1, 0x88, 0x6A, 0xD8, 0xBC, 0x32,
+    0xEC, 0xDA, 0xF1, 0xBC, 0xAF, 0x16, 0x90, 0xCD, 0xEE, 0x5F, 0x92, 0x0B,
+    0xCE, 0xB9, 0x26, 0xCF, 0x18, 0xAE, 0x8C, 0x9B, 0x06, 0x0B, 0x83, 0x4D,
+    0x99, 0x31, 0x98, 0x3B, 0x29, 0xE1, 0x16, 0x6A, 0xA4, 0x5E, 0xE8, 0x10,
+    0x5F, 0x5B, 0x72, 0x3A, 0xA1, 0xD9, 0x89, 0x70, 0x61, 0xD9, 0xC2, 0x25,
+    0x53, 0x5C, 0x44, 0x10, 0x27, 0xD7, 0xF2, 0x68, 0x75, 0x3F, 0xA3, 0xA7,
+    0xCF, 0x02, 0x03, 0x49, 0xB4, 0xE4, 0xAF, 0x08, 0xEA, 0xAE, 0x97, 0x07,
+    0x36, 0xC8, 0xD5, 0x24, 0xC6, 0x51, 0x8B, 0x91, 0x9A, 0x14, 0x91, 0x67,
+    0x6A, 0xC0, 0xC3, 0x0E, 0x7C, 0xD8, 0x1F, 0xD2, 0x31, 0x07, 0x59, 0x5D,
+    0x1D, 0xBD, 0x8E, 0xAE, 0xD7, 0x01, 0xBA, 0xDE, 0x0B, 0xDA, 0xA6, 0xBC,
+    0x9A, 0xD1, 0x39, 0x59, 0x8F, 0xE5, 0x72, 0x65, 0x0F, 0x2A, 0x2D, 0x90,
+    0x56, 0xE9, 0xDA, 0xF5, 0x4A, 0x26, 0xD3, 0xB3, 0x56, 0x19, 0x84, 0x00,
+    0x3A, 0x11, 0x78, 0x83,
   };
-  static unsigned char dh1024_g[] = {
+  static const uint8_t dh2048_g[] = {
     0x02,
   };
   DH *dh;
+
   if ((dh = DH_new()) == NULL)
     return NULL;
-  if (dh->p) {
-    BN_free(dh->p);
-    dh->p = NULL;
-  }
-  if (dh->g) {
-    BN_free(dh->g);
-    dh->g = NULL;
+
+#ifdef HAVE_DH_SET0_PQG
+  BIGNUM *p, *g;
+  p = BN_bin2bn(dh2048_p, sizeof(dh2048_p), NULL);
+  if (!p) {
+    lock_file(stderr);
+    fprintf(stderr, "%s Error in BN_bin2bn 1!\n", time_string());
+    unlock_file(stderr);
+    DH_free(dh);
+    return NULL;
   }
 
-  dh->p = BN_bin2bn(dh1024_p, sizeof(dh1024_p), NULL);
+  g = BN_bin2bn(dh2048_g, sizeof(dh2048_g), NULL);
+  if (!g) {
+    lock_file(stderr);
+    fprintf(stderr, "%s Error in BN_bin2bn 2!\n", time_string());
+    unlock_file(stderr);
+    BN_free(p);
+    DH_free(dh);
+    return NULL;
+  }
+
+  DH_set0_pqg(dh, p, NULL, g);
+#else
+  dh->p = BN_bin2bn(dh2048_p, sizeof(dh2048_p), NULL);
   if (!dh->p) {
-    do_rawlog(LT_ERR, "Error in BN_bin2bn 1");
+    lock_file(stderr);
+    fprintf(stderr, "%s Error in BN_bin2bn 1!\n", time_string());
+    unlock_file(stderr);
     DH_free(dh);
     return NULL;
   }
 
-  dh->g = BN_bin2bn(dh1024_g, sizeof(dh1024_g), NULL);
+  dh->g = BN_bin2bn(dh2048_g, sizeof(dh2048_g), NULL);
   if (!dh->g) {
-    do_rawlog(LT_ERR, "ERror in BN_bin2bn 2");
+    lock_file(stderr);
+    fprintf(stderr, "%s Error in BN_bin2bn 2!\n", time_string());
+    unlock_file(stderr);
     DH_free(dh);
     return NULL;
   }
+#endif
 
   return dh;
 }
 
+SSL *
+ssl_alloc_struct(void)
+{
+  return SSL_new(ctx);
+}
+
 /** Associate an SSL object with a socket and return it.
  * \param sock socket descriptor to associate with an SSL object.
  * \return pointer to SSL object.
@@ -292,7 +482,7 @@ ssl_setup_socket(int sock)
   SSL *ssl;
   BIO *bio;
 
-  ssl = SSL_new(ctx);
+  ssl = ssl_alloc_struct();
   bio = BIO_new_socket(sock, BIO_NOCLOSE);
   BIO_set_nbio(bio, 1);
   SSL_set_bio(ssl, bio, bio);
@@ -307,14 +497,12 @@ ssl_setup_socket(int sock)
  * be better.
  */
 void
-ssl_close_connection(SSL * ssl)
+ssl_close_connection(SSL *ssl)
 {
   SSL_shutdown(ssl);
   SSL_free(ssl);
 }
 
-
-
 /** Given an accepted connection on the listening socket, set up SSL.
  * \param sock an accepted socket (returned by accept())
  * \param state pointer to place to return connection state.
@@ -345,7 +533,6 @@ ssl_resume(int sock, int *state)
   return ssl;
 }
 
-
 /** Perform an SSL handshake.
  * In some cases, a handshake may block, so we may have to call this
  * function again. Accordingly, we return state information that
@@ -354,14 +541,13 @@ ssl_resume(int sock, int *state)
  * \return resulting state of the object.
  */
 int
-ssl_handshake(SSL * ssl)
+ssl_handshake(SSL *ssl)
 {
   int ret;
   int state = 0;
-  int err;
 
   if ((ret = SSL_do_handshake(ssl)) <= 0) {
-    switch (err = SSL_get_error(ssl, ret)) {
+    switch (SSL_get_error(ssl, ret)) {
     case SSL_ERROR_WANT_READ:
       /* We must want for the socket to be readable, and then repeat
        * the call.
@@ -387,7 +573,7 @@ ssl_handshake(SSL * ssl)
   return state;
 }
 
-/** Given connection state, determine if an SSL_accept needs to be 
+/** Given connection state, determine if an SSL_accept needs to be
  * performed. This is a just a wrapper so we don't have to expose
  * our internal state management stuff.
  * \param state an ssl connection state.
@@ -399,7 +585,7 @@ ssl_need_accept(int state)
   return (state & MYSSL_ACCEPT);
 }
 
-/** Given connection state, determine if an SSL_handshake needs to be 
+/** Given connection state, determine if an SSL_handshake needs to be
  * performed. This is a just a wrapper so we don't have to expose
  * our internal state management stuff.
  * \param state an ssl connection state.
@@ -428,7 +614,7 @@ ssl_want_write(int state)
  * \return ssl state flags indicating success, pending, or failure.
  */
 int
-ssl_accept(SSL * ssl)
+ssl_accept(SSL *ssl)
 {
   int ret;
   int state = 0;
@@ -462,7 +648,10 @@ ssl_accept(SSL * ssl)
       if (SSL_get_verify_result(ssl) == X509_V_OK) {
         /* The client sent a certificate which verified OK */
         X509_NAME_oneline(X509_get_subject_name(peer), buf, 256);
-        do_log(LT_CONN, 0, 0, "SSL client certificate accepted: %s", buf);
+        lock_file(stderr);
+        fprintf(stderr, "%s SSL client certificate accepted: %s", time_string(),
+                buf);
+        unlock_file(stderr);
         state |= MYSSL_VERIFIED;
       }
     }
@@ -471,7 +660,6 @@ ssl_accept(SSL * ssl)
   return state;
 }
 
-
 /** Given an SSL object and its last known state, attempt to read from it.
  * \param ssl pointer to SSL object.
  * \param state saved state of SSL object.
@@ -483,7 +671,7 @@ ssl_accept(SSL * ssl)
  * \return new state of SSL object, or -1 if the connection closed.
  */
 int
-ssl_read(SSL * ssl, int state, int net_read_ready, int net_write_ready,
+ssl_read(SSL *ssl, int state, int net_read_ready, int net_write_ready,
          char *buf, int bufsize, int *bytes_read)
 {
   if ((net_read_ready && !(state & MYSSL_WBOR)) ||
@@ -531,17 +719,17 @@ ssl_read(SSL * ssl, int state, int net_read_ready, int net_write_ready,
  * \return new state of SSL object, or -1 if the connection closed.
  */
 int
-ssl_write(SSL * ssl, int state, int net_read_ready, int net_write_ready,
-          unsigned char *buf, int bufsize, int *offset)
+ssl_write(SSL *ssl, int state, int net_read_ready, int net_write_ready,
+          const char *buf, int bufsize, int *offset)
 {
   int r;
-  if ((net_write_ready && bufsize) || (net_read_ready && !(state & MYSSL_WBOR))) {
+  if ((net_write_ready && bufsize) ||
+      (net_read_ready && !(state & MYSSL_WBOR))) {
     state &= ~(MYSSL_WBOR | MYSSL_WB);
     r = SSL_write(ssl, buf + *offset, bufsize);
     switch (SSL_get_error(ssl, r)) {
     case SSL_ERROR_NONE:
       /* We wrote something, but maybe not all */
-      bufsize -= r;
       *offset += r;
       break;
     case SSL_ERROR_WANT_WRITE:
@@ -564,67 +752,11 @@ ssl_write(SSL * ssl, int state, int net_read_ready, int net_write_ready,
   return state;
 }
 
-
 static void
 ssl_errordump(const char *msg)
 {
-  fprintf(stderr, "%s\n", msg);
+  lock_file(stderr);
+  fprintf(stderr, "%s %s\n", time_string(), msg);
   ERR_print_errors(bio_err);
+  unlock_file(stderr);
 }
-
-#ifdef BROKEN
-
-/* The below were various attempts to serialize and save/restore 
- * SSL objects. It ain't pretty, and it don't work.
- */
-
-void
-ssl_write_session(FILE * fp, SSL * ssl)
-{
-  SSL_SESSION *s;
-  s = SSL_get_session(ssl);
-  PEM_write_SSL_SESSION(fp, s);
-}
-
-void
-ssl_read_session(FILE * fp)
-{
-  SSL_SESSION s;
-  PEM_read_SSL_SESSION(fp, &s, NULL, NULL);
-  SSL_CTX_add_session(ctx, &s);
-}
-
-void
-ssl_write_ssl(FILE * fp, SSL * ssl)
-{
-  BIO *bio;
-  SSL_CIPHER *cipher;
-  bio = SSL_get_rbio(ssl);
-  cipher = SSL_CIPHER_get_current_cipher(ssl);
-  fwrite(bio, sizeof(BIO), 1, fp);
-  fwrite(ssl->version, sizeof(ssl->version), 1, fp);
-  fwrite(ssl->type, sizeof(ssl->type), 1, fp);
-  fwrite(ssl->rwstate, sizeof(ssl->type), 1, fp);
-  fwrite(ssl->rstate, sizeof(ssl->type), 1, fp);
-  fwrite(ssl->state, sizeof(ssl->type), 1, fp);
-  fwrite(cipher, sizeof(cipher), 1, fp);
-}
-
-SSL *
-ssl_read_ssl(FILE * fp, int sock)
-{
-  SSL *ssl;
-  BIO *bio;
-
-  bio = BIO_new(BIO_s_socket());
-  fread(bio, sizeof(BIO), 1, fp);
-  ssl = SSL_new(ctx);
-  fread(ssl, sizeof(SSL), 1, fp);
-  SSL_set_ssl_method(ssl, SSLv23_server_method());
-  SSL_set_bio(ssl, bio, bio);
-  return ssl;
-}
-#endif                          /* BROKEN */
-
-
-#endif                          /* HAS_OPENSSL */
