@@ -6,16 +6,101 @@
  *
  */
 
-#include "config.h"
 #include <signal.h>
+#include <stdio.h>
+#include <errno.h>
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+#ifdef HAVE_SYS_EVENTFD_H
+#include <sys/eventfd.h>
+#endif
+#include <fcntl.h>
+#include <unistd.h>
+
 #include "conf.h"
-#include "externs.h"
-#include "confmagic.h"
+#include "mysocket.h"
+#include "sig.h"
+#include "log.h"
+
+int sigrecv_fd = -1;
+int signotifier_fd = -1;
 
 #ifndef HAVE_SIGPROCMASK
 static Sigfunc saved_handlers[NSIG];
 #endif
 
+#ifndef WIN32
+/* Since we install restartable signal handler calls, we have to have a way to
+ * tell the
+ * main game loop that a signal has been received. Use a pipe, or on linux, an
+ * eventfd. */
+
+/** Set up signal notification pipeline. Should only be called once. */
+void
+sigrecv_setup(void)
+{
+#ifdef HAVE_EVENTFD
+  sigrecv_fd = signotifier_fd = eventfd(0, EFD_CLOEXEC | EFD_NONBLOCK);
+  if (sigrecv_fd < 0)
+    penn_perror("sigrecv_setup: eventfd");
+#else
+  int fds[2];
+
+#ifdef HAVE_PIPE2
+  if (pipe2(fds, O_CLOEXEC | O_NONBLOCK) < 0) {
+    penn_perror("sigrecv_setup: pipe2");
+    return;
+  }
+
+  sigrecv_fd = fds[0];
+  signotifier_fd = fds[1];
+
+#else
+  int flags;
+
+  if (pipe(fds) < 0) {
+    penn_perror("sigrecv_setup: pipe");
+    return;
+  }
+  sigrecv_fd = fds[0];
+  signotifier_fd = fds[1];
+  set_close_exec(sigrecv_fd);
+  make_nonblocking(sigrecv_fd);
+  set_close_exec(signotifier_fd);
+  make_nonblocking(signotifier_fd);
+#endif
+#endif
+}
+
+/** Called by signal handler functions to announce a signal has been
+ * received. */
+void
+sigrecv_notify(void)
+{
+  int64_t data = 1;
+  if (write(signotifier_fd, &data, sizeof data) < 0) {
+    penn_perror("sigrecv_notify: write");
+  }
+}
+
+/** Called by shovechars() to acknowledge that a signal has been received.
+ */
+void
+sigrecv_ack(void)
+{
+  int64_t data;
+  if (read(sigrecv_fd, &data, sizeof data) < 0) {
+    if (errno != EAGAIN)
+      penn_perror("sigrecv_ack: read");
+  }
+}
+
+#endif
+
 /** Our own version of signal().
  * We're going to rewrite the signal() function in terms of
  * sigaction, where available, to ensure consistent semantics.
@@ -42,7 +127,7 @@ install_sig_handler(int signo, Sigfunc func)
   if (sigaction(signo, &act, &oact) < 0)
     return SIG_ERR;
   return oact.sa_handler;
-#else                           /* No sigaction, drat. */
+#else /* No sigaction, drat. */
   return signal(signo, func);
 #endif
 }
@@ -52,8 +137,8 @@ install_sig_handler(int signo, Sigfunc func)
  * \param func signal handler function to reload.
  */
 void
-reload_sig_handler(int signo __attribute__ ((__unused__)),
-                   Sigfunc func __attribute__ ((__unused__)))
+reload_sig_handler(int signo __attribute__((__unused__)),
+                   Sigfunc func __attribute__((__unused__)))
 {
 #if !(defined(HAVE_SIGACTION) || defined(SIGNALS_KEPT))
   signal(signo, func);
@@ -72,7 +157,7 @@ ignore_signal(int signo)
   sigemptyset(&act.sa_mask);
   act.sa_flags = 0;
   sigaction(signo, &act, NULL);
-#else                           /* No sigaction, drat. */
+#else /* No sigaction, drat. */
   signal(signo, SIG_IGN);
 #endif
 }
