@@ -7,7 +7,6 @@
  */
 
 #include "copyrite.h"
-#include "config.h"
 
 #include <string.h>
 #ifdef HAVE_INTTYPES_H
@@ -15,20 +14,18 @@
 #endif
 #include <stdio.h>
 
+#include "ansi.h"
+#include "attrib.h"
 #include "conf.h"
-#include "externs.h"
-#include "mushdb.h"
 #include "dbdefs.h"
+#include "externs.h"
 #include "flags.h"
 #include "lock.h"
-#include "attrib.h"
-#include "ansi.h"
-#include "pueblo.h"
+#include "mushdb.h"
+#include "notify.h"
 #include "parse.h"
-#include "confmagic.h"
-
-/* Hack added by Thorvald for object_header Pueblo */
-static int couldunparse;
+#include "pueblo.h"
+#include "strutil.h"
 
 /** Format an object's name (and dbref and flags).
  * This is a wrapper for real_unparse() that conditionally applies
@@ -39,19 +36,9 @@ static int couldunparse;
  * \return static formatted object name string.
  */
 const char *
-unparse_object(dbref player, dbref loc)
+unparse_object(dbref player, dbref loc, int an_flag)
 {
-  static PUEBLOBUFF;
-  const char *result;
-  result = real_unparse(player, loc, 0, 0, 0);
-  if (couldunparse) {
-    PUSE;
-    tag_wrap("A", tprintf("XCH_CMD=\"examine #%d\"", loc), result);
-    PEND;
-    return pbuff;
-  } else {
-    return result;
-  }
+  return real_unparse(player, loc, 0, 0, 0, an_flag, NULL);
 }
 
 /** Format an object's name, obeying MYOPIC/ownership rules.
@@ -63,19 +50,9 @@ unparse_object(dbref player, dbref loc)
  * \return static formatted object name string.
  */
 const char *
-unparse_object_myopic(dbref player, dbref loc)
+unparse_object_myopic(dbref player, dbref loc, int an_flag)
 {
-  static PUEBLOBUFF;
-  const char *result;
-  result = real_unparse(player, loc, 1, 0, 1);
-  if (couldunparse) {
-    PUSE;
-    tag_wrap("A", tprintf("XCH_CMD=\"examine #%d\"", loc), result);
-    PEND;
-    return pbuff;
-  } else {
-    return result;
-  }
+  return real_unparse(player, loc, 1, 0, 1, an_flag, NULL);
 }
 
 /** Format an object's name, obeying MYOPIC/ownership and NAMEFORMAT.
@@ -87,22 +64,13 @@ unparse_object_myopic(dbref player, dbref loc)
  * \endverbatim
  * \param player the looker.
  * \param loc the object being looked at.
+ * \param pe_info pe_info used for evaluating NAMEFORMAT
  * \return static formatted object name string.
  */
 const char *
-unparse_room(dbref player, dbref loc)
+unparse_room(dbref player, dbref loc, NEW_PE_INFO *pe_info)
 {
-  static PUEBLOBUFF;
-  const char *result;
-  result = real_unparse(player, loc, 1, 1, 1);
-  if (couldunparse) {
-    PUSE;
-    tag_wrap("A", tprintf("XCH_CMD=\"examine #%d\"", loc), result);
-    PEND;
-    return pbuff;
-  } else {
-    return result;
-  }
+  return real_unparse(player, loc, 1, 1, 1, AN_LOOK, pe_info);
 }
 
 /** Format an object's name in several ways.
@@ -111,21 +79,24 @@ unparse_room(dbref player, dbref loc)
  * a nameformat or nameaccent if present.
  * \param player the looker.
  * \param loc dbref of the object being looked at.
- * \param obey_myopic if 0, always show Name(#xxxFLAGS); if 1, don't
+ * \param obey_myopic if 0, always show Name(\#xxxFLAGS); if 1, don't
  * do so if player is MYOPIC or doesn't own loc.
  * \param use_nameformat if 1, apply a NAMEFORMAT attribute if available.
  * \param use_nameaccent if 1, apply a NAMEACCENT attribute if available.
+ * \param an_flags AN_* flag to determine whether to use monikerd/ansid name
+ * \param pe_info pe_info used for evaluating NAMEFORMAT
  * \return address of a static buffer containing the formatted name.
  */
 const char *
 real_unparse(dbref player, dbref loc, int obey_myopic, int use_nameformat,
-             int use_nameaccent)
+             int use_nameaccent, int an_flags, NEW_PE_INFO *pe_info)
 {
   static char buf[BUFFER_LEN], *bp;
   static char tbuf1[BUFFER_LEN];
+  static PUEBLOBUFF;
   char *p;
+  bool monikered = 0;
 
-  couldunparse = 0;
   if (!(GoodObject(loc) || (loc == NOTHING) || (loc == AMBIGUOUS) ||
         (loc == HOME)))
     return T("*NOTHING*");
@@ -138,30 +109,28 @@ real_unparse(dbref player, dbref loc, int obey_myopic, int use_nameformat,
     return T("*HOME*");
   default:
     if (use_nameaccent)
-      strcpy(tbuf1, accented_name(loc));
+      strcpy(tbuf1, AaName(loc, an_flags, &monikered));
     else
-      strcpy(tbuf1, Name(loc));
+      strcpy(tbuf1, AName(loc, an_flags, &monikered));
     if (IsExit(loc) && obey_myopic) {
       if ((p = strchr(tbuf1, ';')))
         *p = '\0';
     }
-    if ((Can_Examine(player, loc) || can_link_to(player, loc) ||
+    if ((Can_Examine(player, loc) || can_link_to(player, loc, pe_info) ||
          JumpOk(loc) || ChownOk(loc) || DestOk(loc)) &&
         (!Myopic(player) || !obey_myopic)) {
       /* show everything */
-      if (SUPPORT_PUEBLO)
-        couldunparse = 1;
       bp = buf;
-      if (ANSI_NAMES && ShowAnsi(player))
-        safe_format(buf, &bp, "%s%s%s(#%d%s)", ANSI_HILITE, tbuf1,
-                    ANSI_END, loc, unparse_flags(loc, player));
+      if (ANSI_NAMES && !monikered)
+        safe_format(buf, &bp, "%s%s%s(#%d%s)", ANSI_HILITE, tbuf1, ANSI_END,
+                    loc, unparse_flags(loc, player));
       else
         safe_format(buf, &bp, "%s(#%d%s)", tbuf1, loc,
                     unparse_flags(loc, player));
       *bp = '\0';
     } else {
       /* show only the name */
-      if (ANSI_NAMES && ShowAnsi(player)) {
+      if (ANSI_NAMES && !monikered) {
         bp = buf;
         safe_format(buf, &bp, "%s%s%s", ANSI_HILITE, tbuf1, ANSI_END);
         *bp = '\0';
@@ -172,66 +141,57 @@ real_unparse(dbref player, dbref loc, int obey_myopic, int use_nameformat,
   /* buf now contains the default formatting of the name. If we
    * have @nameaccent, though, we might change to that.
    */
-  if (use_nameformat && nameformat(player, loc, tbuf1, buf))
-    return tbuf1;
+  if (use_nameformat && nameformat(player, loc, tbuf1, buf, 0, pe_info))
+    p = tbuf1;
   else
-    return buf;
+    p = buf;
+
+  if (SUPPORT_HTML) {
+    char tmp[100];
+    PUSE;
+    snprintf(tmp, sizeof tmp, "XCH_CMD=\"look #%d\"", loc);
+    tag_wrap("A", tmp, p);
+    PEND;
+    return pbuff;
+  } else {
+    return p;
+  }
 }
 
 /** Build the name of loc as seen by a player inside it, but only
  * if it has a NAMEFORMAT.
- * This function needs to avoid using a static buffer, so pass in 
+ * This function needs to avoid using a static buffer, so pass in
  * a pointer to an allocated BUFFER_LEN array.
  * \param player the looker.
  * \param loc dbref of location being looked at.
  * \param tbuf1 address to store formatted name of loc.
  * \param defname the name as it would be formatted without NAMEFORMAT.
+ * \param localize should q-registers be preserved when evaluating the attr?
+ * \param pe_info pe_info used for evaluating NAMEFORMAT
  * \retval 1 a NAMEFORMAT was found, and tbuf1 contains formatted name.
  * \retval 0 no NAMEFORMAT on loc, tbuf1 is undefined.
  */
 int
-nameformat(dbref player, dbref loc, char *tbuf1, char *defname)
+nameformat(dbref player, dbref loc, char *tbuf1, char *defname, bool localize,
+           NEW_PE_INFO *pe_info)
 {
-  ATTR *a;
-  char *wsave[10], *rsave[NUMQ];
-  char *arg, *bp, *arg2;
-  char const *sp, *save;
+  ufun_attrib ufun;
+  PE_REGS *pe_regs;
+  int flags = UFUN_IGNORE_PERMS | UFUN_REQUIRE_ATTR;
 
-  int j;
-  a = atr_get(loc, "NAMEFORMAT");
-  if (a) {
-    arg = (char *) mush_malloc(BUFFER_LEN, "string");
-    arg2 = (char *) mush_malloc(BUFFER_LEN, "string");
-    if (!arg)
-      mush_panic("Unable to allocate memory in nameformat");
-    save_global_regs("nameformat", rsave);
-    for (j = 0; j < 10; j++) {
-      wsave[j] = global_eval_context.wenv[j];
-      global_eval_context.wenv[j] = NULL;
-    }
-    for (j = 0; j < NUMQ; j++)
-      global_eval_context.renv[j][0] = '\0';
-    strcpy(arg, unparse_dbref(loc));
-    global_eval_context.wenv[0] = arg;
-    strcpy(arg2, defname);
-    global_eval_context.wenv[1] = arg2;
-    sp = save = safe_atr_value(a);
-    bp = tbuf1;
-    process_expression(tbuf1, &bp, &sp, loc, player, player,
-                       PE_DEFAULT, PT_DEFAULT, NULL);
-    *bp = '\0';
-    free((Malloc_t) save);
-    for (j = 0; j < 10; j++) {
-      global_eval_context.wenv[j] = wsave[j];
-    }
-    restore_global_regs("nameformat", rsave);
-    mush_free((Malloc_t) arg, "string");
-    mush_free((Malloc_t) arg2, "string");
-    return 1;
-  } else {
-    /* No @nameformat attribute */
+  if (localize)
+    flags |= UFUN_LOCALIZE;
+
+  if (!fetch_ufun_attrib("NAMEFORMAT", loc, &ufun, flags))
     return 0;
-  }
+
+  pe_regs = pe_regs_create(PE_REGS_ARG, "nameformat");
+  pe_regs_setenv(pe_regs, 0, unparse_dbref(loc));
+  pe_regs_setenv_nocopy(pe_regs, 1, defname);
+
+  call_ufun(&ufun, tbuf1, player, player, pe_info, pe_regs);
+  pe_regs_free(pe_regs);
+  return 1;
 }
 
 /** Give a string representation of a dbref.
@@ -276,10 +236,10 @@ unparse_uinteger(uintmax_t num)
 {
   static char str[128];
 #ifndef PRIuMAX
-  /* Probably not right */
-#define PRIuMAX "lld"
+/* Probably not right */
+#define PRIuMAX "llu"
 #endif
-  sprintf(str, "%" PRIuMAX, num);
+  snprintf(str, sizeof str, "%" PRIuMAX, num);
   return str;
 }
 
@@ -290,9 +250,10 @@ unparse_uinteger(uintmax_t num)
 char *
 unparse_number(NVAL num)
 {
-  static char str[100];         /* Should be large enough for even the HUGE floats */
+  /* 100 is NOT large enough for even the huge floats */
+  static char str[1000]; /* Should be large enough for even the HUGE floats */
   char *p;
-  sprintf(str, "%.*f", FLOAT_PRECISION, num);
+  snprintf(str, 1000, "%.*f", FLOAT_PRECISION, num);
 
   if ((p = strchr(str, '.'))) {
     p += strlen(p);
