@@ -9,16 +9,18 @@
  * BUILDDATE is correct
  * \endverbatim
  */
-#include "config.h"
+
 #include "copyrite.h"
-#include "conf.h"
-#include "externs.h"
 #include "version.h"
-#include "patches.h"
+
+#include "conf.h"
+#include "notify.h"
+#include "strutil.h"
+#include "gitinfo.h"
+
 #ifndef WIN32
 #include "buildinf.h"
 #endif
-#include "confmagic.h"
 
 void do_version(dbref player);
 
@@ -28,14 +30,15 @@ void do_version(dbref player);
 void
 do_version(dbref player)
 {
-
   notify_format(player, T("You are connected to %s"), MUDNAME);
+  if (MUDURL && *MUDURL)
+    notify_format(player, T("Address: %s"), MUDURL);
   notify_format(player, T("Last restarted: %s"),
                 show_time(globals.start_time, 0));
-  notify_format(player, "PennMUSH version %s patchlevel %s %s", VERSION,
+  notify_format(player, T("PennMUSH version %s patchlevel %s %s"), VERSION,
                 PATCHLEVEL, PATCHDATE);
-#ifdef PATCHES
-  notify_format(player, "Patches: %s", PATCHES);
+#ifdef GIT_REVISION
+  notify_format(player, T("Git revision: %s"), GIT_REVISION);
 #endif
 #ifdef WIN32
   notify_format(player, T("Build date: %s"), __DATE__);
