@@ -1,11 +1,33 @@
 package PennMUSH;
-
+use strict;
+use warnings;
 use File::Copy;
 use File::Path;
 use MUSHConnection;
-
+use POSIX qw/:sys_wait_h/;
+use feature qw/say/;
+no if $] >= 5.017011, warnings => 'experimental::smartmatch'; # Don't blow up on Perls older than 5.18
 my @pids = ();
 
+$SIG{"CHLD"} = sub  {
+  while ((my $child = waitpid(-1, WNOHANG))  > 0) {
+    if ($child ~~ @pids) {
+      my $status = ${^CHILD_ERROR_NATIVE};
+      my $errmsg = "Child PennMUSH process $child exited: ";
+      if (WIFEXITED($status)) {
+        $errmsg .= "Exit code " . WEXITSTATUS($status);
+      } elsif (WIFSIGNALED($status)) {
+        $errmsg .= "Killed with signal " . WTERMSIG($status);
+      }
+      say $errmsg;
+      say "Last lines of log:";
+      say `tail log/netmush.log`;
+      say "Valgrind:";
+      say `cat valgrind-*`;
+    }
+  }
+};
+
 sub new {
   my $proto = shift;
   my $class = ref($proto) || $proto;
@@ -27,7 +49,7 @@ sub start {
   my $self = shift;
   srand();
   $self->{HOST} = "localhost" unless defined $self->{HOST};
-  if (!exists $self->{PORT} || $self->{PORT} <= 0) {      
+  if (!exists $self->{PORT} || $self->{PORT} <= 0) {
       $self->{PORT} = int(rand(2000)) + 12000;
   }
   my $port = $self->{PORT};
@@ -38,6 +60,8 @@ sub start {
              "compress_program" => "",
              "uncompress_program" => "",
              "compress_suffix" => "",
+	     "mem_check" => "yes",
+	     "dict_file" => "",
              @_);
   copy("../game/alias.cnf", "testgame/alias.cnf");
   copy("../game/names.cnf", "testgame/names.cnf");
@@ -56,23 +80,30 @@ sub start {
     my $line;
     push(@pids, $child);
     $self->{PID} = $child;
-    foreach $j (1..20) {
-      next unless open(LOG, "testgame/log/netmush.log");
-      while ($line = <LOG>) {
-        close(LOG), return $port if $line =~ /^Listening on port $port /;
+    sleep 5;
+    foreach $j (1..10) {
+      next unless open my $LOG, "<", "testgame/log/netmush.log";
+      while ($line = <$LOG>) {
+        return $port if $line =~ /Listening on port $port /;
       }
+      close $LOG;
     } continue {
-      sleep(1);
+      sleep ($self->{VALGRIND} ? 10 : 5);
     }
     die "Could not start game process properly; pid $child!\n";
   } elsif (defined($child)) {
     chdir("testgame");
-    my @execargs = ("./netmush", "--no-session", "test.cnf");
-    unshift @execargs, "valgrind", "--tool=memcheck", '--log-file=../valgrind-%p.log', "--leak-check=full"
-	if $self->{VALGRIND};
+    my @execargs = ("./netmush", "--no-session", "--disable-socket-quota", "--tests");
+    if ($self->{VALGRIND}) {
+      unshift @execargs, "valgrind", "--tool=memcheck", '--log-file=../valgrind-%p.log',
+	"--leak-check=full", "--track-origins=yes";
+      push @execargs, "--no-pcre-jit";
+    }
+    push @execargs, "test.cnf";
     exec @execargs;
+    die "Unable to run '@execargs': $!\n";
   } else {
-    die "Could not spawn game process!\n";
+    die "Could not spawn game process: $!\n";
   }
 }
 
@@ -81,19 +112,19 @@ sub copyConfig {
   my $to = shift;
   my %subs = @_;
 
-  open(FROM, "<$from") || die "Could not open template configuration.\n";
-  open(TO, ">$to") || die "Could not write test configuration.\n";
+  open my $FROM, "<", $from or die "Could not open template configuration: $!\n";
+  open my $TO, ">", $to or die "Could not write test configuration: $!\n";
   my $line;
-  while ($line = <FROM>) {
+  while ($line = <$FROM>) {
     next if $line =~ /^\s*#/o;
     next unless $line =~ /^\s*(\w+)\s/o;
     my $key = $1;
     $line = $key . " " . $subs{$key} . "\n" if defined($subs{$key});
   } continue {
-    print TO $line;
+    print $TO $line;
   }
-  close(TO);
-  close(FROM);
+  close $TO;
+  close $FROM;
 }
 
 sub login {
@@ -107,7 +138,7 @@ sub loginGod {
 }
 
 END {
-  kill("TERM", @pids);
+  kill("INT", @pids);
 }
 
 1;
